__all__ = ()

from ...field_parsers import bool_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import (
    bool_optional_putter_factory, nullable_string_optional_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, nullable_string_validator_factory, url_required_validator_factory
)

from .constants import DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX


# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory(
    'description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX
)


# spoiler

parse_spoiler = bool_parser_factory('spoiler', False)
put_spoiler = bool_optional_putter_factory('spoiler', False)
validate_spoiler = bool_validator_factory('spoiler', False)


# url

parse_url = force_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
validate_url = url_required_validator_factory('url')



__all__ = ()

import sys, warnings
from importlib.util import module_from_spec, spec_from_file_location
from os.path import basename as get_file_name, splitext as split_file_name_and_extension
from py_compile import PyCompileError, compile as compile_module

from scarletio import HybridValueDictionary, RichAttributeErrorBaseType, WeakSet, include

from .constants import (
    IN_DIRECTORY_PLUGIN_RP, LOADING_PLUGINS, PLUGINS, PLUGIN_ACTION_FLAG_SYNTAX_CHECK, PLUGIN_STATE_LOADED,
    PLUGIN_STATE_UNDEFINED, PLUGIN_STATE_UNLOADED, PLUGIN_STATE_UNSATISFIED, PLUGIN_STATE_VALUE_TO_NAME
)
from .exceptions import DoNotLoadPlugin, PluginError
from .helpers import PROTECTED_NAMES, _get_path_plugin_name, _validate_entry_or_exit
from .import_overwrite.module_proxy_type import PluginModuleProxyType
from .import_overwrite.module_spec_type import PluginModuleSpecType
from .plugin_root import register_plugin_root
from .snapshot import calculate_snapshot_difference, revert_snapshot, take_snapshot


PLUGIN_LOADER = include('PLUGIN_LOADER')


class Plugin(RichAttributeErrorBaseType):
    """
    Represents a plugin.
    
    Attributes
    ----------
    _added_variable_names : `list of `str`
        A list of the added variables' names to the module.
    _child_plugins : `None`, ``WeakSet`` of ``Plugin``
        Child plugins.
    _default_variables : `None`, `HybridValueDictionary` of (`str`, `object`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _entry_point : `None`, `str`, `callable`
        Internal slot used by the ``.entry_point`` property.
    _exit_point : `None`, `str`, `callable`
        Internal slot used by the ``.exit_point`` property.
    _extend_default_variables : `bool`
        Internal slot used by the ``.extend_default_variables`` property.
    _locked : `bool`
        The internal slot used for the ``.locked`` property.
    _parent_plugins : `None`, ``WeakSet`` of ``Plugin``
        Parent plugins.
    _snapshot_difference : `None`, `list` of ``BaseSnapshotType``
        Snapshot difference if applicable. Defaults to `None`.
    _snapshot_extractions : `None`, `list` of `list` of ``BaseSnapshotType``
        Additional snapshots to extract from own.
    _spec : ``PluginModuleSpecType``
        The module specification for the plugin's module's import system related state.
    _state : `int`
        The state of the plugin. Can be:
        +-------------------------------+-------+
        | Respective name               | Value |
        +===============================+=======+
        | PLUGIN_STATE_UNDEFINED        | 0     |
        +-------------------------------+-------+
        | PLUGIN_STATE_LOADED           | 1     |
        +-------------------------------+-------+
        | PLUGIN_STATE_UNLOADED         | 2     |
        +-------------------------------+-------+
        | PLUGIN_STATE_UNSATISFIED      | 3     |
        +-------------------------------+-------+
    _sub_module_plugins : `None`, ``WeakSet`` of ``Plugin``
        Sub module plugins.
    _take_snapshot : `bool`
        Whether snapshot difference should be taken.
    """
    __slots__ = (
        '__weakref__', '_added_variable_names', '_child_plugins', '_default_variables', '_entry_point',
        '_exit_point', '_extend_default_variables', '_locked', '_parent_plugins', '_snapshot_difference',
        '_snapshot_extractions', '_spec', '_state', '_sub_module_plugins', '_take_snapshot'
    )
    
    def __new__(
        cls, name, path, entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference,
        default_variables
    ):
        """
        Creates a plugin with the given parameters. If a plugin already exists with the given name, returns
        that.
        
        Parameters
        ----------
        name : `None`, `str`
            The plugin's name (or import path).
        path : `str`
            Path to the plugin file.
        entry_point : `None`, `str`, `callable`
            The entry point of the plugin.
        exit_point : `None`, `str`, `callable`
            The exit point of the plugin.
        extend_default_variables : `bool`
            Whether the plugin should use the loader's default variables or just it's own.
        locked : `bool`
            Whether the plugin should be picked up by the `{}_all` methods of the plugin loader.
        take_snapshot_difference: `bool`
            Whether snapshots should be taken before and after loading a plugin, and when the plugin is unloaded,
            the snapshot difference should be reverted.
        default_variables : `None`, `HybridValueDictionary` of (`str`, `object`) items
            An optionally weak value dictionary to store objects for assigning them to modules before loading them.
            If would be empty, is set as `None` instead.
        
        Returns
        -------
        self : ``Plugin``
        
        Raises
        ------
        ModuleNotFoundError
            If the plugin was not found.
        """
        if (name is None):
            name = _get_path_plugin_name(path)
        
        try:
            self = PLUGINS[name]
        except KeyError:
            pass
        else:
            if (default_variables is not None):
                self.add_default_variables(**default_variables)
            
            return self
        
        spec = spec_from_file_location(name, path)
        if spec is None:
            raise ModuleNotFoundError(name)
        
        spec = PluginModuleSpecType(spec)
        register_plugin_root(name)
        
        self = object.__new__(cls)
        self._added_variable_names = []
        self._child_plugins = None
        self._parent_plugins = None
        self._default_variables = default_variables
        self._entry_point = entry_point
        self._exit_point = exit_point
        self._extend_default_variables = extend_default_variables
        self._locked = locked
        self._snapshot_difference = None
        self._snapshot_extractions = None
        self._spec = spec
        self._state = PLUGIN_STATE_UNDEFINED
        self._sub_module_plugins = None
        self._take_snapshot = take_snapshot_difference
        
        PLUGINS[name] = self
        
        return self
    
    
    def __hash__(self):
        """Returns the plugin's ``._spec``'s `.origin`'s hash."""
        return hash(self._spec.origin)
    
    
    def __repr__(self):
        """Returns the plugin's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        repr_parts.append(' name = ')
        repr_parts.append(repr(self._spec.name))
        
        state = self._state
        repr_parts.append(', state = ')
        state_name = PLUGIN_STATE_VALUE_TO_NAME.get(state, '???')
        repr_parts.append(state_name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(state))
        
        if self._locked:
            repr_parts.append(', locked = True')
        
        default_variables = self._default_variables
        if self._extend_default_variables:
            if (default_variables is not None):
                repr_parts.append(' extends defaults with: ')
                repr_parts.append(repr(default_variables))
        else:
            if default_variables is None:
                repr_parts.append(' clears defaults')
            else:
                repr_parts.append(' clears defaults and uses: ')
                repr_parts.append(repr(default_variables))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.sort_key > other.sort_key
    
    
    def __lt__(self, other):
        """Returns whether self is less than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.sort_key < other.sort_key
    
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the plugin.
        
        Parameters
        ----------
        **variables : Keyword Parameters
            Variables to assigned to the plugin's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        if not variables:
            return
        
        default_variables = self._default_variables
        if default_variables is None:
            default_variables = HybridValueDictionary()
            self._default_variables = default_variables
        
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(
                    f'The passed {key!r} is a protected variable name of module type.'
                )
            default_variables[key] = value
    
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the plugin.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        if default_variables is None:
            return
        
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
        
        if default_variables:
            return
        
        self._default_variables = None
    
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the plugin.
        """
        self._default_variables = None
    
    
    @property
    def entry_point(self):
        """
        Get-set-del descriptor for modifying the plugin's entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._entry_point
    
    
    @entry_point.setter
    def entry_point(self, entry_point):
        if not _validate_entry_or_exit(entry_point):
            raise TypeError(
                f'`{self.__class__.__name__}.entry_point` can be `None`, `str`, `callable`, got '
                f'{entry_point.__class__.__name__}; {entry_point!r}.'
            )
        
        self._entry_point = entry_point
    
    
    @entry_point.deleter
    def entry_point(self):
        self._entry_point = None
    
    
    @property
    def exit_point(self):
        """
        Get-set-del descriptor for modifying the plugin's exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._exit_point
    
    @exit_point.setter
    def exit_point(self, exit_point):
        if not _validate_entry_or_exit(exit_point):
            raise TypeError(
                f'`{self.__class__.__name__}.exit_point` can be `None`, `str`, `callable`, got '
                f'{exit_point.__class__.__name__}; {exit_point!r}.'
            )
        
        self._exit_point = exit_point
    
    @exit_point.deleter
    def exit_point(self):
        self._exit_point = None
    
    
    @property
    def extend_default_variables(self):
        """
        Get-set descriptor to define whether the plugin uses the loader's default variables or just it's own.
        
        Accepts and returns `bool`.
        """
        return self._extend_default_variables
    
    
    @extend_default_variables.setter
    def extend_default_variables(self, extend_default_variables):
        extend_default_variables_type = extend_default_variables.__class__
        if extend_default_variables_type is bool:
            pass
        elif issubclass(extend_default_variables_type, int):
            extend_default_variables = bool(extend_default_variables)
        else:
            raise TypeError(
                f'`extend_default_variables` can be `bool`, got '
                f'{extend_default_variables_type.__name__}; {extend_default_variables!r}.'
            )
        
        self._extend_default_variables = extend_default_variables
    
    
    @property
    def locked(self):
        """
        Get-set property to define whether the plugin should be picked up by the `{}_all` methods of the
        plugin loader.
        
        Accepts and returns `bool`.
        """
        return self._locked
    
    
    @locked.setter
    def locked(self, locked):
        locked_type = locked.__class__
        if locked_type is bool:
            pass
        elif issubclass(locked_type, int):
            locked = bool(locked)
        else:
            raise TypeError(
                f'`locked` can be `bool`, got {locked_type.__name__}; {locked!r}.'
            )
        
        self._locked = locked
    
    
    def _load(self):
        """
        Loads the module and returns it. If it is already loaded returns `None`.
        
        Returns
        -------
        module : `None`, `ModuleType`
        
        Raises
        ------
        ImportError
            Circular imports.
        RuntimeError
            Module already imported.
        BaseException
            Any exception raised from the plugin file.
        """
        if self in LOADING_PLUGINS:
            raise ImportError(
                f'{self.name} is already being loaded. This can happen if there are circular imports. The following '
                f'plugins are being loaded right now: '
                f'{", ".join(plugin.name for plugin in LOADING_PLUGINS)}'
            )
        
        try:
            LOADING_PLUGINS.add(self)
            state = self._state
            if state == PLUGIN_STATE_UNDEFINED:
                
                spec = self._spec
                
                maybe_proxy_module = sys.modules.get(spec.name, None)
                if (maybe_proxy_module is not None) and (not isinstance(maybe_proxy_module, PluginModuleProxyType)):
                    raise RuntimeError(
                        f'Module `{spec.name}` is already imported.'
                    )
                
                module = spec.get_module()
                
                if (module is None):
                    module_from_spec(spec)
                    
                    module = spec.get_module()
                    loaded = self._load_module()
                
                else:
                    loaded = True
                
                if loaded:
                    state = PLUGIN_STATE_LOADED
                else:
                    state = PLUGIN_STATE_UNSATISFIED
                    module = None
                
                self._state = state
                return module
            
            if state == PLUGIN_STATE_LOADED:
                # return None -> already loaded
                return None
            
            if state in (PLUGIN_STATE_UNLOADED, PLUGIN_STATE_UNSATISFIED):
                # reload
                module_from_spec(self._spec)
                loaded = self._load_module()
                
                if loaded:
                    state = PLUGIN_STATE_LOADED
                    module =  self._spec.get_module()
                else:
                    state = PLUGIN_STATE_UNSATISFIED
                    module = None
                
                self._state = state
                
                return module
            
            return None
            # no more cases
        
        finally:
            LOADING_PLUGINS.discard(self)
    
    
    def _load_module(self):
        """
        Loads the module and returns whether it was successfully loaded.
        
        Returns
        -------
        loaded : `bool`
        
        Raises
        ------
        BaseException
            Any exception raised from the plugin file.
        """
        spec = self._spec
        module = spec.get_module()
        
        added_variable_names = self._added_variable_names
        if self._extend_default_variables:
            for name, value in PLUGIN_LOADER._default_variables.items():
                setattr(module, name, value)
                added_variable_names.append(name)
        
        default_variables = self._default_variables
        if (default_variables is not None):
            for name, value in default_variables.items():
                setattr(module, name, value)
                added_variable_names.append(name)
        
        for plugin in self.iter_loaded_plugins_in_directory():
            setattr(module, plugin.short_name, plugin.get_module_proxy())
        
        active_plugins_at_start = LOADING_PLUGINS.copy()
        
        if self._take_snapshot:
            snapshot_old = take_snapshot()
        else:
            snapshot_old = None
        
        try:
            spec.loader.exec_module(module)
        except DoNotLoadPlugin:
            loaded = False
        else:
            loaded = True
        
        if loaded:
            active_plugins_at_end = LOADING_PLUGINS.copy()
            plugin_intersection = active_plugins_at_start & active_plugins_at_end
            plugin_intersection.discard(self)
            
            if (snapshot_old is not None):
                snapshot_new = take_snapshot()
                snapshot_difference = calculate_snapshot_difference(snapshot_new, snapshot_old)
                
                for snapshot_extraction in self.iter_snapshot_extractions():
                    snapshot_difference = calculate_snapshot_difference(
                        snapshot_difference, snapshot_extraction
                    )
                
                self.clear_snapshot_extractions()
                
                self._snapshot_difference = snapshot_difference
                
                for plugin in plugin_intersection:
                    plugin.add_snapshot_extraction(snapshot_difference)
            
            else:
                for snapshot_extraction in self.iter_snapshot_extractions():
                    for plugin in plugin_intersection:
                        plugin.add_snapshot_extraction(snapshot_extraction)
                
                self.clear_snapshot_extractions()
        
        return loaded
    
    
    def _check_for_syntax(self):
        """
        Checks the file's new syntax.
        
        Used when reloading, to avoid unloading un-loadable files.
        
        This method is blocking. Run it inside of an executor.
        
        Returns
        -------
        exception : PluginError
            PluginError wrapping invalid syntax.
        """
        if (self._state == PLUGIN_STATE_LOADED) and self._spec.is_initialised():
            file_name = self.file_name
            # python files might not be `.py` files, which we should not compile.
            if file_name.endswith('.py'):
                try:
                    compile_module(file_name, doraise = True)
                except FileNotFoundError:
                    # If the file is deleted, is fine.
                    pass
                
                except FileExistsError:
                    pass
                
                except PyCompileError as err:
                    exception_value = err.exc_value
                    if isinstance(exception_value, SyntaxError):
                        cause = exception_value
                    else:
                        cause = err
                    
                    try:
                        raise PluginError(action = PLUGIN_ACTION_FLAG_SYNTAX_CHECK, cause = cause, plugin = self)
                    finally:
                        cause = None
    
    
    def _unload(self):
        """
        Unloads the module and returns it. If it is already unloaded returns `None`.
        
        Returns
        -------
        module : `None`, `ModuleType`
        
        Raises
        ------
        SyntaxError
            If `check_for_syntax` and the file's syntax is incorrect.
        """
        state = self._state
        if state == PLUGIN_STATE_UNDEFINED:
            return None
        
        if state == PLUGIN_STATE_LOADED:
            self.clear_child_plugins()
            self.clear_parent_plugins()
            
            snapshot_difference = self._snapshot_difference
            if (snapshot_difference is not None):
                self._snapshot_difference = None
                revert_snapshot(snapshot_difference)
            
            self._state = PLUGIN_STATE_UNLOADED
            
            spec = self._spec
            try:
                del sys.modules[spec.name]
            except KeyError:
                pass
            
            return spec.get_module()
        
        if state in (PLUGIN_STATE_UNLOADED, PLUGIN_STATE_UNSATISFIED):
            return None
        
        # no more cases
    
    def _unassign_variables(self):
        """
        Unassigns the assigned variables to the respective module and clears ``._added_variable_names``.
        """
        module = self._spec.get_module()
        if module is None:
            return
        
        added_variable_names = self._added_variable_names
        for name in added_variable_names:
            try:
                delattr(module, name)
            except AttributeError:
                pass
        
        added_variable_names.clear()
    
    
    @property
    def name(self):
        """
        Returns the plugin's name.
        
        Returns
        -------
        name : `str`
        """
        return self._spec.name
    
    
    @property
    def path(self):
        """
        Returns the plugin's name.
        
        Returns
        -------
        name : `str`
        """
        return self._spec.origin
    
    
    @property
    def short_name(self):
        """
        Returns the plugin's name's shortened version.
        
        Returns
        -------
        name : `str`
        """
        name = self._spec.name
        dot_index = name.rfind('.')
        if dot_index == -1:
            short_name = None
        else:
            short_name = name[dot_index + 1:]
        
        return short_name
    
    
    @property
    def file_name(self):
        """
        Returns the plugin's file's name.
        
        Returns
        -------
        file_name : `str`
        """
        return self._spec.origin
    
    
    @property
    def sort_key(self):
        """
        Returns the plugin's sort key.
        
        Returns
        -------
        sort_key : `tuple` of `str`
        """
        return self._spec.name.split('.')
    
    
    def is_loaded(self):
        """
        Returns whether the plugin is loaded.
        
        Returns
        -------
        is_loaded : `bool`
        """
        return (self._state == PLUGIN_STATE_LOADED)
    
   
    def is_unsatisfied(self):
        """
        Returns whether the plugin is unsatisfied.
        
        Returns
        -------
        is_unsatisfied : `bool`
        """
        return (self._state == PLUGIN_STATE_UNSATISFIED)
    
    
    def _unlink(self):
        """
        Removes the plugin's module from the loaded ones.
        
        Should not be called on loaded plugins.
        """
        state = self._state
        if state == PLUGIN_STATE_UNDEFINED:
            pass
        
        elif state == PLUGIN_STATE_LOADED:
            self._state = PLUGIN_STATE_UNLOADED
            module = self._spec.get_module()
            
            if self._extend_default_variables:
                for name in PLUGIN_LOADER._default_variables:
                    try:
                        delattr(module, name)
                    except AttributeError:
                        pass
            
            default_variables = self._default_variables
            if (default_variables is not None):
                for name in default_variables:
                    delattr(module, name)
            
            try:
                del sys.modules[self._spec.name]
            except KeyError:
                pass
        
        elif state in (PLUGIN_STATE_UNLOADED, PLUGIN_STATE_UNSATISFIED):
            try:
                del sys.modules[self._spec.name]
            except KeyError:
                pass
        
        # no more cases
        
        plugins_by_name = PLUGIN_LOADER._plugins_by_name
        for name in (self.name, self.short_name, self.path):
            if plugins_by_name.get(name, None) is self:
                try:
                    del plugins_by_name[name]
                except KeyError:
                    pass
    
    
    def add_child_plugin(self, plugin):
        """
        Registers a child plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to register.
        """
        if plugin is self:
            return
        
        child_plugins = self._child_plugins
        if (child_plugins is None):
            child_plugins = WeakSet()
            self._child_plugins = child_plugins
        
        child_plugins.add(plugin)
    
    
    def iter_child_plugins(self):
        """
        Iterates over the child plugins.
        
        This method is an iterable generator.
        
        Yields
        ------
        child_plugin : `None`
        """
        child_plugins = self._child_plugins
        if (child_plugins is not None):
            yield from child_plugins
    
    
    def are_child_plugins_present_in(self, plugins):
        """
        Returns whether all the child plugins are present in the given `plugins`.
        
        Parameters
        ----------
        plugins : `iterable` of ``Plugin``
            Already present plugins to check satisfaction form.
        
        Returns
        -------
        are_child_plugins_present_in : `bool`
        """
        child_plugins = self._child_plugins
        if (child_plugins is None):
            return True
        
        if child_plugins <= plugins:
            return True
        
        return False
    
    
    def clear_child_plugins(self):
        """
        Clears the child plugins of the plugin.
        """
        child_plugins = self._child_plugins
        if (child_plugins is not None):
            self._child_plugins = None
            
            for child_plugin in child_plugins:
                child_plugin.remove_parent_plugin(self)
    
    
    def remove_child_plugin(self, child_plugin):
        """
        Removes the given plugin from the plugin's children.
        
        Parameters
        ----------
        child_plugin : ``Plugin``
            The plugin to remove.
        """
        child_plugins = self._child_plugins
        if (child_plugins is not None):
            try:
                child_plugins.remove(child_plugin)
            except KeyError:
                pass
            else:
                if not child_plugins:
                    self._child_plugins = child_plugins
    
    
    def add_snapshot_extraction(self, snapshots):
        """
        Adds snapshot extraction to the plugin.
        
        Parameters
        ----------
        snapshots : `list` of ``BaseSnapshotType``
        """
        snapshot_extractions = self._snapshot_extractions
        if (snapshot_extractions is None):
            snapshot_extractions = []
            self._snapshot_extractions = snapshot_extractions
        
        snapshot_extractions.append(snapshots)
    
    
    def iter_snapshot_extractions(self):
        """
        Iterates over the snapshot extractions of the plugin.
        
        This method is an iterable generator.
        
        Yields
        ------
        snapshots : `list` of ``BaseSnapshotType``
        """
        snapshot_extractions = self._snapshot_extractions
        if (snapshot_extractions is not None):
            yield from snapshot_extractions
    
    
    def clear_snapshot_extractions(self):
        """
        Clears the snapshot extractions of the plugin.
        """
        self._snapshot_extractions = None
    
    
    def add_parent_plugin(self, plugin):
        """
        Registers a parent plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to register.
        """
        if plugin is self:
            return
        
        parent_plugins = self._parent_plugins
        if (parent_plugins is None):
            parent_plugins = WeakSet()
            self._parent_plugins = parent_plugins
        
        parent_plugins.add(plugin)
    
    
    def iter_parent_plugins(self):
        """
        Iterates over the parent plugins.
        
        This method is an iterable generator.
        
        Yields
        ------
        parent_plugin : `None`
        """
        parent_plugins = self._parent_plugins
        if (parent_plugins is not None):
            yield from parent_plugins
    
    
    def are_parent_plugins_present_in(self, plugins):
        """
        Returns whether all the parent plugins are present in the given `plugins`.
        
        Parameters
        ----------
        plugins : `iterable` of ``Plugin``
            Already present plugins to check satisfaction form.
        
        Returns
        -------
        are_parent_plugins_present_in : `bool`
        """
        parent_plugins = self._parent_plugins
        if (parent_plugins is None):
            return True
        
        if parent_plugins <= plugins:
            return True
        
        return False
    
    
    def clear_parent_plugins(self):
        """
        Clears the parent plugins of the plugin.
        """
        parent_plugins = self._parent_plugins
        if (parent_plugins is not None):
            self._parent_plugins = parent_plugins
            
            for parent_plugin in parent_plugins:
                parent_plugin.remove_child_plugin(self)
    
    
    def remove_parent_plugin(self, parent_plugin):
        """
        Removes the given plugin from the plugin's parents.
        
        Parameters
        ----------
        parent_plugin : ``Plugin``
            The plugin to remove.
        """
        parent_plugins = self._parent_plugins
        if (parent_plugins is not None):
            try:
                parent_plugins.remove(parent_plugin)
            except KeyError:
                pass
            else:
                if not parent_plugins:
                    self._parent_plugins = parent_plugins
    
    
    @property
    def _module(self):
        """
        Deprecated attribute of ``Plugin``.
        """
        warnings.warn(
            f'`{self.__class__.__name__}._module` is deprecated, please use `.get_module()` instead.',
            FutureWarning,
            stacklevel = 2,
        )
        return self._spec.get_module()
    
    
    def get_module(self):
        """
        Returns the module of the plugin.
        
        Returns
        -------
        module : `None`, `ModuleType`
        """
        return self._spec.get_module()
    
    
    def get_module_proxy(self):
        """
        Returns a proxy to the plugin.
        
        Returns
        -------
        module_proxy : ``PluginModuleProxyType``
        """
        return self._spec.get_module_proxy()
    
    
    def is_directory(self):
        """
        Returns whether the plugin is a directory.
        
        Returns
        -------
        is_directory : `bool`
        """
        return split_file_name_and_extension(get_file_name(self._spec.origin))[0] == '__init__'
    
    
    def iter_loaded_plugins_in_directory(self):
        """
        Iterates over the loaded plugins directly under this one. This one must be a directory (so an `__init__` file).
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin : ``Plugin``
        """
        if not self.is_directory():
            return
        
        self_name = self.name
        for plugin in PLUGINS.values():
            if not plugin.is_loaded():
                continue
            
            other_name = plugin.name
            if not other_name.startswith(self_name):
                continue
            
            if IN_DIRECTORY_PLUGIN_RP.fullmatch(other_name[len(self_name):]) is None:
                continue
            
            yield plugin
    
    
    def add_sub_module_plugin(self, plugin):
        """
        Registers a sub module plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to register.
        """
        if plugin is self:
            return
        
        sub_module_plugins = self._sub_module_plugins
        if (sub_module_plugins is None):
            sub_module_plugins = WeakSet()
            self._sub_module_plugins = sub_module_plugins
        
        sub_module_plugins.add(plugin)
    
    
    def iter_sub_module_plugins(self):
        """
        Iterates over the sub module plugins.
        
        This method is an iterable generator.
        
        Yields
        ------
        sub_module_plugin : `None`
        """
        sub_module_plugins = self._sub_module_plugins
        if (sub_module_plugins is not None):
            yield from sub_module_plugins
    
    
    def are_sub_module_plugins_present_in(self, plugins):
        """
        Returns whether all the sub module plugins are present in the given `plugins`.
        
        Parameters
        ----------
        plugins : `iterable` of ``Plugin``
            Already present plugins to check satisfaction form.
        
        Returns
        -------
        are_sub_module_plugins_present_in : `bool`
        """
        sub_module_plugins = self._sub_module_plugins
        if (sub_module_plugins is None):
            return True
        
        if sub_module_plugins <= plugins:
            return True
        
        return False
    
    
    def clear_sub_module_plugins(self):
        """
        Clears the sub module plugins of the plugin.
        """
        sub_module_plugins = self._sub_module_plugins
        if (sub_module_plugins is not None):
            self._sub_module_plugins = None
    
    
    def remove_sub_module_plugin(self, sub_module_plugin):
        """
        Removes the given plugin from the plugin's sub modules.
        
        Parameters
        ----------
        sub_module_plugin : ``Plugin``
            The plugin to remove.
        """
        sub_module_plugins = self._sub_module_plugins
        if (sub_module_plugins is not None):
            try:
                sub_module_plugins.remove(sub_module_plugin)
            except KeyError:
                pass
            else:
                if not sub_module_plugins:
                    self._sub_module_plugins = sub_module_plugins



__all__ = ()

from ...application import ApplicationIntegrationType
from ...application_command.application_command.constants import (
    NAME_LENGTH_MAX as APPLICATION_COMMAND_NAME_LENGTH_MAX, NAME_LENGTH_MIN as APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from ...field_parsers import (
    default_entity_parser_factory, entity_id_parser_factory, nullable_entity_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    default_entity_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    nullable_entity_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    default_entity_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_entity_validator_factory, nullable_string_array_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

# authorizer_user_ids

def parse_authorizer_user_ids(data):
    """
    Parses authorizer user identifiers.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    authorizer_user_ids : `None | dict<ApplicationIntegrationType, int>`
    """
    authorizer_user_ids_data = data.get('authorizing_integration_owners', None)
    if (authorizer_user_ids_data is None) or (not authorizer_user_ids_data):
        return None
    
    authorizer_user_ids = {}
    
    for key, value in authorizer_user_ids_data.items():
        integration_type = ApplicationIntegrationType(ApplicationIntegrationType.VALUE_TYPE(key))
        user_id = int(value)
        authorizer_user_ids[integration_type] = user_id
    
    return authorizer_user_ids


def put_authorizer_user_ids(authorizer_user_ids, data, defaults):
    """
    Serializes the authorizer user identifiers.
    
    Parameters
    ----------
    authorizer_user_ids : `None | dict<ApplicationIntegrationType, int>`
        The identifiers of the authorizer users.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    authorizer_user_ids_data = {}
    
    if (authorizer_user_ids is not None):
        for integration_type, user_id in authorizer_user_ids.items():
            key = str(integration_type.value)
            value = str(user_id)
            authorizer_user_ids_data[key] = value
    
    data['authorizing_integration_owners'] = authorizer_user_ids_data
    return data


def validate_authorizer_user_ids(authorizer_user_ids):
    """
    Validates the authorizer user identifiers.
    
    Parameters
    ----------
    authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int, ClientUserBase | int>`
        Authorizer user identifiers to validate.
    
    Returns
    -------
    authorizer_user_ids : `None | dict<ApplicationIntegrationType, int>`
    
    Raises
    ------
    TypeError
        - Value of invalid type given.
    """
    if authorizer_user_ids is None:
        return None
    
    if not isinstance(authorizer_user_ids, dict):
        raise TypeError(
            f'`authorizer_user_ids` can be `None`,'
            f'`dict<{ApplicationIntegrationType.__name__} | {ApplicationIntegrationType.VALUE_TYPE.__name__}, '
            f'int | {ClientUserBase.__name__}, got '
            f'{type(authorizer_user_ids).__name__}; {authorizer_user_ids!r}.'
        )
    
    if not authorizer_user_ids:
        return None
    
    validated_authorizer_user_ids = {}
    
    for key, value in authorizer_user_ids.items():
        if isinstance(key, ApplicationIntegrationType):
            integration_type = key
        
        elif isinstance(key, ApplicationIntegrationType.VALUE_TYPE):
            integration_type = ApplicationIntegrationType(key)
        
        else:
            raise TypeError(
                f'`authorizer_user_ids` keys can be '
                f'`{ApplicationIntegrationType.__name__}`, `{ApplicationIntegrationType.VALUE_TYPE.__name__}`, '
                f'got {type(key).__name__}; {key!r}; '
                f'authorizer_user_ids = {authorizer_user_ids!r}.'
            )
        
        if isinstance(value, int):
            user_id = value
        
        elif isinstance(value, ClientUserBase):
            user_id = value.id
        
        else:
            raise TypeError(
                f'`authorizer_user_ids` values can be `int`, `{ClientUserBase.__name__}`, '
                f'got {type(value).__name__}; {value!r}; '
                f'authorizer_user_ids = {authorizer_user_ids!r}.'
            )
        
        validated_authorizer_user_ids[integration_type] = user_id
    
    return validated_authorizer_user_ids


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('message_interaction_id')


# interacted_message_id

parse_interacted_message_id = entity_id_parser_factory('interacted_message_id')
put_interacted_message_id = entity_id_optional_putter_factory('interacted_message_id')
validate_interacted_message_id = entity_id_validator_factory('interacted_message_id', NotImplemented, include = 'Message')


# name & sub_command_name_stack

def parse_name_and_sub_command_name_stack(data):
    """
    Parses out the message interaction's name and sub command name stack from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message interaction data.
    
    Returns
    -------
    name : `str`
        Message interaction name.
    sub_command_name_stack : `None`, `tuple` of `str`
        The sub-command-group and sub-command names.
    """
    raw_name = data.get('name', None)
    if raw_name is None:
        name = ''
        sub_command_name_stack = None
    
    else:
        name_split = raw_name.split(' ')
        
        name_split_length = len(name_split)
        if name_split_length == 1:
            name = name_split[0]
            sub_command_name_stack = None
        
        elif name_split_length > 1:
            name = name_split[0]
            sub_command_name_stack = (*name_split[1:],)
        
        else:
            name = ''
            sub_command_name_stack = None
    
    return name, sub_command_name_stack


def put_name_and_sub_command_name_stack(name_and_sub_command_name_stack, data, defaults):
    """
    Puts the message interaction's name and sub command name stack into the given `data` json serializable object.
    
    Parameters
    ----------
    name_and_sub_command_name_stack : `tuple` (`str`, `None` | `tuple` of `str`)
       Message interaction name and sub command name stack.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    name, sub_command_name_stack = name_and_sub_command_name_stack
    if sub_command_name_stack is None:
        raw_name = name
    else:
        raw_name = ' '.join([name, *sub_command_name_stack])
    
    data['name'] = raw_name
    return data


validate_name = force_string_validator_factory(
    'name', APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX
)

validate_sub_command_name_stack = nullable_string_array_validator_factory('sub_command_name_stack', ordered = False)


# response_message_id

parse_response_message_id = entity_id_parser_factory('original_response_message_id')
put_response_message_id = entity_id_optional_putter_factory('original_response_message_id')
validate_response_message_id = entity_id_validator_factory('response_message_id', NotImplemented, include = 'Message')


# target_message_id

parse_target_message_id = entity_id_parser_factory('target_message_id')
put_target_message_id = entity_id_optional_putter_factory('target_message_id')
validate_target_message_id = entity_id_validator_factory('target_message_id', NotImplemented, include = 'Message')


# target_user

parse_target_user = nullable_entity_parser_factory('target_user', User)
put_target_user = nullable_entity_optional_putter_factory(
    'target_user', ClientUserBase, force_include_internals = True
)
validate_target_user = nullable_entity_validator_factory('target_user', ClientUserBase)


# triggering_interaction

parse_triggering_interaction = nullable_entity_parser_factory(
    'triggering_interaction_metadata', NotImplemented, include = 'MessageInteraction'
)
put_triggering_interaction = nullable_entity_optional_putter_factory(
    'triggering_interaction_metadata', NotImplemented, can_include_internals = True, include = 'MessageInteraction',
)
validate_triggering_interaction = nullable_entity_validator_factory(
    'triggering_interaction', NotImplemented, include = 'MessageInteraction'
)


# type

parse_type = preinstanced_parser_factory(
    'type', NotImplemented, NotImplemented, include = 'InteractionType', include_default_attribute_name = 'none'
)

put_type = preinstanced_putter_factory('type')

validate_type = preinstanced_validator_factory('type', NotImplemented, include = 'InteractionType')


# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user = default_entity_putter_factory('user', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)



from .guild_join_request_form_response import *

__all__ = guild_join_request_form_response.__all__

from ....utils.module_deprecation import deprecated_import

deprecated_import(GuildJoinRequestFormResponse, 'GuildRequestFormResponse')



__all__ = ()

from ...field_parsers import force_string_parser_factory
from ...field_putters import force_string_putter_factory
from ...field_validators import force_string_validator_factory

from .constants import ID_LENGTH_MAX, ID_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN

# id

parse_id = force_string_parser_factory('id')
put_id = force_string_putter_factory('id')
validate_id = force_string_validator_factory('integration_account_id', ID_LENGTH_MIN, ID_LENGTH_MAX)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ()

from ...field_parsers import int_parser_factory, nullable_string_parser_factory
from ...field_putters import int_putter_factory, url_optional_putter_factory
from ...field_validators import (
    int_conditional_validator_factory, nullable_string_validator_factory, url_optional_validator_factory
)

from .constants import URL_LENGTH_MAX

# height

parse_height = int_parser_factory('height', 0)
put_height = int_putter_factory('height')
validate_height = int_conditional_validator_factory(
    'height',
    0,
    lambda height : height >= 0,
    '>= 0',
)

# url

parse_url = nullable_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
# url validator doesnt allow attachment:\\image.png formats
validate_url = nullable_string_validator_factory('url', 0, URL_LENGTH_MAX)

# proxy_url

parse_proxy_url = nullable_string_parser_factory('proxy_url')
put_proxy_url = url_optional_putter_factory('proxy_url')
validate_proxy_url = url_optional_validator_factory('proxy_url')

# width

parse_width = int_parser_factory('width', 0)
put_width = int_putter_factory('width')
validate_width = int_conditional_validator_factory(
    'width',
    0,
    lambda width : width >= 0,
    '>= 0',
)



__all__ = ()


# heart beat
HEARTBEAT_TIMEOUT = 20.0
INTERVAL_DEFAULT = 40.0
LATENCY_DEFAULT = 9999.0

# gateway
GATEWAY_CONNECT_TIMEOUT = 30.0
POLL_TIMEOUT = 60.0

# rate limit
GATEWAY_RATE_LIMIT_LIMIT = 120
GATEWAY_RATE_LIMIT_RESET = 60.0

# gateway control
GATEWAY_ACTION_KEEP_GOING = 0
GATEWAY_ACTION_CONNECT = 1
GATEWAY_ACTION_RESUME = 2


# operations / client

GATEWAY_OPERATION_CLIENT_DISPATCH = 0
GATEWAY_OPERATION_CLIENT_HEARTBEAT = 1
GATEWAY_OPERATION_CLIENT_IDENTIFY = 2
GATEWAY_OPERATION_CLIENT_PRESENCE = 3
GATEWAY_OPERATION_CLIENT_VOICE_STATE = 4
GATEWAY_OPERATION_CLIENT_VOICE_PING = 5
GATEWAY_OPERATION_CLIENT_RESUME = 6
GATEWAY_OPERATION_CLIENT_RECONNECT = 7
GATEWAY_OPERATION_CLIENT_REQUEST_GUILD_USERS = 8
GATEWAY_OPERATION_CLIENT_INVALIDATE_SESSION = 9
GATEWAY_OPERATION_CLIENT_HELLO = 10
GATEWAY_OPERATION_CLIENT_HEARTBEAT_ACKNOWLEDGE = 11
GATEWAY_OPERATION_CLIENT_GUILD_SYNC = 12
GATEWAY_OPERATION_CLIENT_REQUEST_SOUNDBOARD_SOUNDS = 31

"""
GATEWAY_OPERATION_CLIENT_DISPATCH : `int` = `0`
    Receive only, used at `._handle_received_message`.
GATEWAY_OPERATION_CLIENT_HEARTBEAT : `int` = `1`
    Send and receive, used at `._beat` and at `._handle_special_operation`.
GATEWAY_OPERATION_CLIENT_IDENTIFY : `int` = `2`
    Send only, used `._identify`.
GATEWAY_OPERATION_CLIENT_PRESENCE : `int` = `3`
    Send only, used at ``Client.edit_presence``.
GATEWAY_OPERATION_CLIENT_VOICE_STATE : `int` = `4`
    Send only, used at `.change_voice_state`.
GATEWAY_OPERATION_CLIENT_VOICE_PING : `int` = `5`
    Removed.
GATEWAY_OPERATION_CLIENT_RESUME : `int` = `6`
    Send only, used at `._resume`.
GATEWAY_OPERATION_CLIENT_RECONNECT : `int` = `7`
    Receive only, used at `._special_operation`.
GATEWAY_OPERATION_CLIENT_REQUEST_GUILD_USERS : `int` = `8`
    Send only, used at ``Client._request_users_loop``, ``Client._request_users`` and at ``Client.request_member``.
GATEWAY_OPERATION_CLIENT_INVALIDATE_SESSION : `int` = `9`
    Receive only, used at `._handle_special_operation`.
GATEWAY_OPERATION_CLIENT_HELLO : `int` = `10`
    Receive only, used at `._handle_special_operation`.
GATEWAY_OPERATION_CLIENT_HEARTBEAT_ACKNOWLEDGE : `int` = `11`
    Receive only, used at `._handle_special_operation`.
GATEWAY_OPERATION_CLIENT_GUILD_SYNC : `int` = `12`
    Send only, not used.
GATEWAY_OPERATION_CLIENT_REQUEST_SOUNDBOARD_SOUNDS : `int` = `31`
    Send only, used to request the guilds' soundboard sounds.
"""

# operations / voice

GATEWAY_OPERATION_VOICE_IDENTIFY = 0
GATEWAY_OPERATION_VOICE_SELECT_PROTOCOL = 1
GATEWAY_OPERATION_VOICE_READY = 2
GATEWAY_OPERATION_VOICE_HEARTBEAT = 3
GATEWAY_OPERATION_VOICE_SESSION_DESCRIPTION = 4
GATEWAY_OPERATION_VOICE_SPEAKING = 5
GATEWAY_OPERATION_VOICE_HEARTBEAT_ACKNOWLEDGE = 6
GATEWAY_OPERATION_VOICE_RESUME = 7
GATEWAY_OPERATION_VOICE_HELLO = 8
GATEWAY_OPERATION_VOICE_RESUMED = 9
# 10 missing
GATEWAY_OPERATION_VOICE_USERS_CONNECT = 11
GATEWAY_OPERATION_VOICE_CLIENT_CONNECT = 12
GATEWAY_OPERATION_VOICE_CLIENT_DISCONNECT = 13
GATEWAY_OPERATION_VOICE_VIDEO_SESSION_DESCRIPTION = 14
GATEWAY_OPERATION_VOICE_VIDEO_SINK = 15
GATEWAY_OPERATION_VOICE_VOICE_BACKEND_VERSION = 16
GATEWAY_OPERATION_VOICE_CHANNEL_OPTIONS_UPDATE = 17
GATEWAY_OPERATION_VOICE_FLAGS = 18
# 19 missing
GATEWAY_OPERATION_VOICE_PLATFORM = 20
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_PREPARE_TRANSITION = 21
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_COMMIT_TRANSITION = 22
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_READY_FOR_TRANSITION = 23
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_PREPARE_EPOCH = 24
GATEWAY_OPERATION_VOICE_MLS_EXTERNAL_SENDER_PACKAGE = 25
GATEWAY_OPERATION_VOICE_MLS_KEY_PACKAGE = 26
GATEWAY_OPERATION_VOICE_MLS_PROPOSALS = 27
GATEWAY_OPERATION_VOICE_MLS_COMMIT_WELCOME = 28
GATEWAY_OPERATION_VOICE_MLS_COMMIT_TRANSITION = 29
GATEWAY_OPERATION_VOICE_MLS_WELCOME = 30
GATEWAY_OPERATION_VOICE_MLS_MLS_INVALID_COMMIT_WELCOME = 31


"""
GATEWAY_OPERATION_VOICE_IDENTIFY : `int` = `0`
    Send only, used at `._identify`.
GATEWAY_OPERATION_VOICE_SELECT_PROTOCOL : `int` = `1`
    Send only, used at `._select_protocol`.
GATEWAY_OPERATION_VOICE_READY : `int` = `2`
    Receive only, used at `._initial_connection`.
GATEWAY_OPERATION_VOICE_HEARTBEAT : `int` = `3`
    Send only, used at `._beat`.
GATEWAY_OPERATION_VOICE_SESSION_DESCRIPTION : `int` = `4`
    Receive only, used at `._received_message`.
GATEWAY_OPERATION_VOICE_SPEAKING : `int` = `5`
    Send and receive, used at `.set_speaking` and at `._handle_operation_speaking`.
GATEWAY_OPERATION_VOICE_HEARTBEAT_ACKNOWLEDGE : `int` = `6`
    Receive only, used at `._handle_operation_heartbeat_acknowledge`.
GATEWAY_OPERATION_VOICE_RESUME : `int` = `7`
    Send only, used at `._resume`.
GATEWAY_OPERATION_VOICE_HELLO : `int` = `8`
    Receive only, used at `._handle_operation_hello`.
GATEWAY_OPERATION_VOICE_RESUMED : `int` = `9`
    Receive only, used at `._handle_operation_resumed`.
GATEWAY_OPERATION_VOICE_USERS_CONNECT : `int` = `11`
    Receive only, used at `._handle_operation_users_connect`.
GATEWAY_OPERATION_VOICE_CLIENT_CONNECT : `int` = `12`
    Receive only, used at `._handle_operation_client_connect`.
GATEWAY_OPERATION_VOICE_CLIENT_DISCONNECT : `int` = `13`
    Receive only, used at `._handle_operation_client_disconnect`.
GATEWAY_OPERATION_VOICE_VIDEO_SESSION_DESCRIPTION : `int` = `14`
    Receive only. Not used.
GATEWAY_OPERATION_VOICE_VIDEO_SINK : `int` = `15`
    Receive and send, not used.
GATEWAY_OPERATION_VOICE_VOICE_BACKEND_VERSION : `int` = `16`
    Unused.
GATEWAY_OPERATION_VOICE_CHANNEL_OPTIONS_UPDATE : `int` = `17`
    Unused.
GATEWAY_OPERATION_VOICE_FLAGS : `int` = `18`
    Receive only, not used.
GATEWAY_OPERATION_VOICE_PLATFORM : `int` = `20`
    Receive only, not used.
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_PREPARE_TRANSITION : `int` = `21`
    Unused.
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_COMMIT_TRANSITION : `int` = `22`
    Unused.
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_READY_FOR_TRANSITION : `int` = `23`
    Unused.
GATEWAY_OPERATION_VOICE_SECURE_FRAMES_PREPARE_EPOCH : `int` = `24`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_EXTERNAL_SENDER_PACKAGE : `int` = `25`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_KEY_PACKAGE : `int` = `26`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_PROPOSALS : `int` = `27`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_COMMIT_WELCOME : `int` = `28`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_COMMIT_TRANSITION : `int` = `29`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_WELCOME : `int` = `30`
    Unused.
GATEWAY_OPERATION_VOICE_MLS_MLS_INVALID_COMMIT_WELCOME : `int` = `31`
    Unused.
"""



__all__ = ()

from .flag_descriptors import (
    FlagBitDescriptor, FlagBitDescriptorDeprecated, FlagBitDescriptorReversed, FlagBitDescriptorReversedDeprecated,
    FlagDescriptor
)


NotImplementedType = type(NotImplemented)


def _ensure_int_in_type_parents(type_parents):
    """
    Ensures that parent 0 is `int` subclass.
    
    Parameters
    ----------
    type_parents : `tuple<type>`
        The type's parents to create.
    
    Returns
    -------
    type_parents : `tuple<type>`
    """
    if not type_parents:
        type_parents = (int,)
    elif not issubclass(type_parents[0], int):
        type_parents = (int, *type_parents)
    
    return type_parents


def _collect_shifts_from_type_parents(meta_type, type_parents, accumulated_shifts):
    """
    Collect the existing shifts from type parents.
    
    Parameters
    ----------
    meta_type : `type`
        The meta type to allow its children to be collected from.
    
    type_parents : `tuple<type>`
        The type's parents to create.
    
    accumulated_shifts : `set<(str, int, None | FlagDeprecation>)>`
        The accumulated shifts by their name.
    """
    for type_parent in type_parents:
        if isinstance(type_parent, meta_type):
            for name, shift in type_parent.__shifts__.items():
                accumulated_shifts.add((name, shift, None))
            
            for name, (shift, deprecation) in type_parent.__deprecated_shifts__.items():
                accumulated_shifts.add((name, shift, deprecation))


def _collect_shifts_from_type_attributes(
    type_attributes, accumulated_shifts, accumulated_flag_descriptors
):
    """
    Collects the new shifts from the type's new attributes.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    accumulated_shifts : `set<(str, int, None | FlagDeprecation)>`
        The accumulated shifts by their name.
    
    accumulated_flag_descriptors : `set<(str, FlagDescriptor)>`
        Flag descriptors.
    """
    for attribute_name, attribute_value in type_attributes.items():
        if isinstance(attribute_value, FlagDescriptor):
            accumulated_flag_descriptors.add((attribute_name, attribute_value))
            accumulated_shifts.add((attribute_name, attribute_value.shift, attribute_value.deprecation))


def _get_shift_name_overlap(accumulated_shifts):
    """
    Collects the keys that are duplicates.
    
    Parameters
    ----------
    accumulated_shifts : `set<(str, int, None | FlagDeprecation)>`
        The accumulated shifts by their name.
    
    Returns
    -------
    duplicates : `set<str>`
    """
    singles = set()
    duplicates = None
    
    # Get shifts by their name. And set them in a dictionary.
    for name, shift, deprecation in accumulated_shifts:
        if name not in singles:
            singles.add(name)
            continue
        
        # Add name to duplicates.
        if duplicates is None:
            duplicates = set()
        
        duplicates.add(name)
    
    return duplicates


def _filter_shifts_by_name(accumulated_shifts, duplicates):
    """
    Filters the shifts for the given names.
    
    Parameters
    ----------
    accumulated_shifts : `set<(str, int, None | FlagDeprecation)>`
        The accumulated shifts by their name.
    
    duplicates : `set<str>`
        Duplicated names.
    
    Returns
    -------
    shifts_by_name : `dict<str, set<int>>`
    """
    shifts_by_name = {name: set() for name in duplicates}
    
    for name, shift, deprecation in accumulated_shifts:
        try:
            shifts = shifts_by_name[name]
        except KeyError:
            continue
        
        shifts.add(shift)
    
    return shifts_by_name


def _build_duplicate_names_exception_message(shifts_by_name):
    """
    Builds exception message for the case when there are duplicate shifts for a single name.
    
    Parameters
    ----------
    shifts_by_name : `dict<str, set<int>>`
        The duplicate shifts under the same name.
    
    Returns
    -------
    exception_message : `str`
    """
    output_parts = []
    field_added = False
    
    for name, shifts in sorted(shifts_by_name.items()):
        if field_added:
            output_parts.append(' ')
        else:
            field_added = True
        
        output_parts.append('Name: `')
        output_parts.append(name)
        output_parts.append('` has duplicate shifts: ')
        
        shift_added = False
        for shift in sorted(shifts):
            if shift_added:
                output_parts.append(', ')
            else:
                shift_added = True
            
            output_parts.append(str(shift))
        
        output_parts.append('.')
    
    return ''.join(output_parts)


def _check_shifts_overlap(accumulated_shifts):
    """
    Raises an exception if there are duplicate shifts under the same name.
    
    Parameters
    ----------
    accumulated_shifts : `set<(str, int, None | FlagDeprecation)>`
        The accumulated shifts by their name.
    
    Raises
    ------
    ValueError
    """
    duplicates = _get_shift_name_overlap(accumulated_shifts)
    if duplicates is None:
        return
    
    raise ValueError(
        _build_duplicate_names_exception_message(_filter_shifts_by_name(accumulated_shifts, duplicates))
    )


def _collect_reverse_descriptors_from_type_parents(meta_type, type_parents, accumulated_reverse_descriptors):
    """
    Collects the `__reverse_descriptors__` fields from the given type parents.
    
    Parameters
    ----------
    meta_type : `type`
        The meta type to allow its children to be collected from.
    
    type_parents : `tuple<type>`
        The parent types of the type to create.
    
    accumulated_reverse_descriptors : `set<(str, int)>`
        Accumulated `__reverse_descriptors__` values.
    """
    for type_parent in type_parents:
        if isinstance(type_parent, meta_type):
            accumulated_reverse_descriptors.add((type_parent.__name__, type_parent.__reverse_descriptors__))


def _check_reverse_descriptors_type(type_name, value):
    """
    Checks whether `__reverse_descriptors__`'s type is acceptable.
    
    Parameters
    ----------
    type_name : `str`
        The type's name.
    
    value : `object`
        The value to check.
    
    Raises
    ------
    TypeError
    """
    if not isinstance(value, bool):
        raise TypeError(
            f'`{type_name}.__reverse_descriptors__` is not `bool`, got `{type(value).__name__}`; {value!r}.'
        )


def _is_reverse_descriptors_contradiction(accumulated_reverse_descriptors):
    """
    Returns whether there is a contradiction in `accumulated_reverse_descriptors` values.
    
    Parameters
    ----------
    accumulated_reverse_descriptors : `set<(str, bool)>`
        Accumulated `__reverse_descriptors__` values.
    
    Returns
    -------
    is_reverse_descriptors_contradiction : `bool`
    """
    iterator = iter(accumulated_reverse_descriptors)
    while True:
        item = next(iterator, None)
        
        # No more items left.
        if item is None:
            return False
        
        stated_value = item[1]
        # Ignore item if `-1`.
        if stated_value == -1:
            continue
        
        break
    
    for item in iterator:
        value = item[1]
        # Ignore item if `-1`.
        if value == -1:
            continue
        
        if value == stated_value:
            continue
        
        return True
    
    return False


def _build_reverse_descriptors_contradiction_exception_message(accumulated_reverse_descriptors):
    """
    Builds a `__reverse_descriptors__` contradiction error message.
    
    Parameters
    ----------
    accumulated_reverse_descriptors : `set<(str, bool)>`
        Accumulated `__reverse_descriptors__` values.
    
    Returns
    -------
    error_message : `str`
    """
    output_parts = ['There is contradiction in `__reverse_descriptors__` values: ']
    
    item_added = False
    
    for type_name, reverse_descriptors in sorted(accumulated_reverse_descriptors):
        if item_added:
            output_parts.append(', ')
        else:
            item_added = True
        
        output_parts.append(type_name)
        output_parts.append(' -> ')
        output_parts.append(repr(reverse_descriptors))
    
    output_parts.append('.')
    return ''.join(output_parts)


def _check_reverse_descriptors_contradiction(accumulated_reverse_descriptors):
    """
    Checks whether there is contradiction in `__reverse_descriptors__` values.
    
    Parameters
    ----------
    accumulated_reverse_descriptors : `set<(str, bool)>`
        Accumulated `__reverse_descriptors__` values.
    
    Raises
    ------
    ValueError
    """
    if _is_reverse_descriptors_contradiction(accumulated_reverse_descriptors):
        raise ValueError(_build_reverse_descriptors_contradiction_exception_message(accumulated_reverse_descriptors))


def _get_reverse_descriptors_value(accumulated_reverse_descriptors):
    """
    Gets what `__reverse_descriptors__` value should the created type use.
    
    Parameters
    ----------
    accumulated_reverse_descriptors : `set<(str, int)>`
        Accumulated `__reverse_descriptors__` values.
    
    Returns
    -------
    reverse_descriptors : `int`
    """
    if accumulated_reverse_descriptors:
        reverse_descriptors = next(iter(accumulated_reverse_descriptors))[1]
    else:
        reverse_descriptors = -1
    return reverse_descriptors


def _build_shifts(accumulated_shifts):
    """
    Builds a new `__shifts__` dictionary.
    
    Parameters
    ----------
    accumulated_shifts : `set<(str, int, None | FlagDeprecation)>`
        The accumulated shifts by their name.
    
    Returns
    -------
    shifts : `dict<str, int>`
    deprecated_shifts : `dict<str, (int, FlagDeprecation)>`
    """
    shifts = {}
    deprecated_shifts = {}
    
    for name, shift, deprecation in accumulated_shifts:
        if deprecation is None:
            shifts[name] = shift
        else:
            deprecated_shifts[name] = (shift, deprecation)
    
    return shifts, deprecated_shifts


def _build_flag_descriptors(type_name, reverse_descriptors, accumulated_flag_descriptors):
    """
    Builds flag descriptors.
    
    Parameters
    ----------
    type_name : `str`
        The created class's name.
    
    reverse_descriptors : `bool`
        Whether the descriptors should be reversed.
    
    accumulated_flag_descriptors : `set<(str, FlagDescriptor)>`
        Flag descriptors.
    
    Returns
    -------
    flag_descriptors : `list<FlagBitDescriptor>`
    """
    if reverse_descriptors:
        descriptor_type = FlagBitDescriptorReversed
        descriptor_type_deprecated = FlagBitDescriptorReversedDeprecated
    else:
        descriptor_type = FlagBitDescriptor
        descriptor_type_deprecated = FlagBitDescriptorDeprecated
        
    flag_descriptors = []
    
    for flag_name, flag_descriptor in accumulated_flag_descriptors:
        deprecation = flag_descriptor.deprecation
        shift = flag_descriptor.shift
        
        if deprecation is None:
            descriptor = descriptor_type(shift, type_name, flag_name)
        else:
            descriptor = descriptor_type_deprecated(shift, type_name, flag_name, deprecation)
        
        flag_descriptors.append(descriptor)
    
    return flag_descriptors


class FlagMeta(type):
    """
    Metatype for bit-flags.
    """
    def __new__(cls, type_name, type_parents, type_attributes, *, reverse_descriptors = ...):
        """
        Creates a new bitwise flag type.
        
        Parameters
        ----------
        type_name : `str`
            The created class's name.
        
        type_parents : `tuple<type>`
            The parent types of the type to create.
        
        type_attributes : `dict<str, object>`
            The type attributes of the created type.
        
        reverse_descriptors : `bool`, Optional (Keyword only)
            Whether 
        
        Returns
        -------
        type : `instance<cls>`
        
        Raises
        ------
        TypeError
            When any requirements are not satisfied.
        """
        type_parents = _ensure_int_in_type_parents(type_parents)
        
        # Assert if type not updated / incorrectly formed
        if __debug__:
            try:
                keys = type_attributes['__keys__']
            except KeyError:
                pass
            else:
                if isinstance(keys, (dict, NotImplementedType)):
                    raise AssertionError('Support for `__keys__` removed.')
        
        assert '__deprecated_keys__' not in type_attributes, 'Support for `__deprecated_keys__` removed.'
        
        # This should not be defined.
        assert '__shifts__' not in type_attributes, '`__shifts__` not allowed to be defined.'
        assert '__shifts_ordered__' not in type_attributes, '`__shifts_ordered__` not allowed to be defined.'
        assert '__deprecated_shifts__' not in type_attributes, '`__deprecated_shifts__` not allowed to be defined.'
        assert '__slots__' not in type_attributes, '`__slots__` not allowed to be defined.'
        
        # Accumulate shifts
        accumulated_shifts = set()
        accumulated_flag_descriptors = set()
        _collect_shifts_from_type_parents(cls, type_parents, accumulated_shifts)
        _collect_shifts_from_type_attributes(
            type_attributes, accumulated_shifts, accumulated_flag_descriptors
        )
        _check_shifts_overlap(accumulated_shifts)
        
        # check whether we reverse
        accumulated_reverse_descriptors = set()
        if (reverse_descriptors is ...):
            reverse_descriptors = -1
        else:
            _check_reverse_descriptors_type(type_name, reverse_descriptors)
            reverse_descriptors = 1 if reverse_descriptors else 0
        
        accumulated_reverse_descriptors.add((type_name, reverse_descriptors))
        
        _collect_reverse_descriptors_from_type_parents(cls, type_parents, accumulated_reverse_descriptors)
        _check_reverse_descriptors_contradiction(accumulated_reverse_descriptors)
        
        # Build new values
        reverse_descriptors = _get_reverse_descriptors_value(accumulated_reverse_descriptors)
        # allow -1 only if we have no descriptors.
        if (reverse_descriptors == -1) and accumulated_flag_descriptors:
            reverse_descriptors = 0
        
        shifts, deprecated_shifts = _build_shifts(accumulated_shifts)
        flag_descriptors = _build_flag_descriptors(type_name, reverse_descriptors, accumulated_flag_descriptors)
        shifts_ordered = sorted(shifts.items())
        
        # Populate attributes
        type_attributes['__reverse_descriptors__'] = reverse_descriptors
        type_attributes['__shifts__'] = shifts
        type_attributes['__shifts_ordered__'] = shifts_ordered
        type_attributes['__deprecated_shifts__'] = deprecated_shifts
        type_attributes.update((flag_descriptor.flag_name, flag_descriptor) for flag_descriptor in flag_descriptors)
        type_attributes['__slots__'] = ()
        
        return type.__new__(cls, type_name, type_parents, type_attributes)



__all__ = ('start_auto_reloader',)

from scarletio import copy_docs

from ..plugin_auto_reloader import PLUGIN_AUTO_RELOADER_MANAGER, PluginAutoReloaderManager


@copy_docs(PluginAutoReloaderManager.start)
def start_auto_reloader():
    return PLUGIN_AUTO_RELOADER_MANAGER.start()



from .module_proxy_type import *
from .module_spec_type import *
from .plugin_finder import *
from .source_loader import *
from .spec_finder_helpers import *
from .tools import *
from .utils import *

__all__ = (
    *module_proxy_type.__all__,
    *module_spec_type.__all__,
    *plugin_finder.__all__,
    *source_loader.__all__,
    *spec_finder_helpers.__all__,
    *tools.__all__,
    *utils.__all__,
)


from .tools import set_spec_finder
set_spec_finder()



__all__ = ('ReactionAddEvent',)

from scarletio import Task, copy_docs

from ...bases import EventBase
from ...core import KOKORO
from ...permission.permission import PERMISSION_MASK_MANAGE_MESSAGES

from ..reaction import Reaction, ReactionType

from .fields import (
    parse_emoji, parse_message, parse_type, parse_user, put_emoji, put_message, put_type, put_user,
    validate_emoji, validate_message, validate_type, validate_user
)
from .helpers import _delete_reaction_with_task


class ReactionAddEvent(EventBase):
    """
    Represents a processed `MESSAGE_REACTION_ADD` dispatch event.
    
    Attributes
    ----------
    emoji : ``Emoji``
        The emoji used as reaction.
    message : ``Message``
        The message on what the reaction is added.
    type : ``ReactionType``
        The reaction's type.
    user : ``ClientUserBase``
        The user who added the reaction.
    
    Class Attributes
    ----------------
    DELETE_REACTION_OK : `int` = `0`
        Returned by ``.delete_reaction_with`` when the client has permission to execute the reaction remove.
    DELETE_REACTION_PERM : `int` = `1`
        Returned by ``.delete_reaction_with`` when the client has no permission to execute the reaction remove.
    DELETE_REACTION_NOT_ADDED : `int` = `2`
        Returned by ``.delete_reaction_with`` when the client has permission to execute the reaction remove, but
        it cannot, because the reaction is not added on the respective message. Not applicable for
        ``ReactionAddEvent``.
    """
    DELETE_REACTION_OK = 0
    DELETE_REACTION_PERM = 1
    DELETE_REACTION_NOT_ADDED = 2
    
    __slots__ = ('emoji', 'message', 'type', 'user')
    
    def __new__(cls, message, emoji, user, *, reaction_type = ...):
        """
        Creates a new reaction add (or delete) instance.
        
        Parameters
        ----------
        message : ``Message``
            The respective message.
        emoji : ``Emoji``
            The emoji used.
        user : ``ClientUserBase``
            The user who reacted.
        reaction_type : ``ReactionType``, `int`, Optional (Keyword only)
            The reaction's type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        emoji = validate_emoji(emoji)
        
        # message
        message = validate_message(message)
        
        # reaction_type
        if reaction_type is ...:
            reaction_type = ReactionType.standard
        else:
            reaction_type = validate_type(reaction_type)
        
        # user
        user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.message = message
        self.emoji = emoji
        self.type = reaction_type
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new reaction add (or delete) event instance.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Reaction event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji = parse_emoji(data)
        self.message = message = parse_message(data)
        self.type = parse_type(data)
        self.user = parse_user(data, message.guild_id)
        return self
    
    
    @classmethod
    def from_fields(cls, message, emoji, user, reaction_type):
        """
        Creates a new reaction add (or delete) instance.
        
        Parameters
        ----------
        message : ``Message``
            The respective message.
        emoji : ``Emoji``
            The emoji used.
        user : ``ClientUserBase``
            The user who reacted.
        reaction_type : ``ReactionType``
            The reaction's type.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji = emoji
        self.message = message
        self.type = reaction_type
        self.user = user
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the reaction add (or delete) event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_emoji(self.emoji, data, defaults)
        message = self.message
        put_message(message, data, defaults)
        put_type(self.type, data, defaults)
        put_user(self.user, data, defaults, guild_id = message.guild_id)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' message = ')
        repr_parts.append(repr(self.message))
        
        repr_parts.append(', emoji = ')
        repr_parts.append(repr(self.emoji))
        
        reaction_type = self.type
        if reaction_type is not ReactionType.standard:
            repr_parts.append(', type = ')
            repr_parts.append(reaction_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(reaction_type.value))
        
        repr_parts.append(', user = ')
        repr_parts.append(repr(self.user))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.message
        yield self.reaction
        yield self.user
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.emoji is not other.emoji:
            return False
        
        if self.message is not other.message:
            return False
        
        if self.type is not other.type:
            return False
        
        if self.user is not other.user:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # emoji
        hash_value ^= hash(self.emoji)
        
        # message
        hash_value ^= hash(self.message)
        
        # typer
        hash_value ^= hash(self.type)
        
        # user
        hash_value ^= self.user.id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the reaction add (or remove) event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji = self.emoji
        new.message = self.message
        new.type = self.type
        new.user = self.user
        return new
    
    
    def copy_with(self, *, emoji = ..., reaction_type = ..., message = ..., user = ...):
        """
        Copies the reaction add (or remove) event with the given fields.
        
        Parameters
        ----------
        emoji : ``Emoji``, Optional (Keyword only)
            The emoji used.
        reaction_type : ``ReactionType``, Optional (Keyword only)
            The reaction's type.
        message : ``Message``, Optional (Keyword only)
            The respective message.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who reacted.
        
        Returns
        -------
        new : `instance<type<self>>`
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # message
        if message is ...:
            message = self.message
        else:
            message = validate_message(message)
        
        # type
        if reaction_type is ...:
            reaction_type = self.type
        else:
            reaction_type = validate_type(reaction_type)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        new = object.__new__(type(self))
        new.emoji = emoji
        new.message = message
        new.type = reaction_type
        new.user = user
        return new
    
    
    @property
    def reaction(self):
        """
        Returns the reaction of the reaction add event.
        
        Returns
        -------
        reaction : ``Reaction``
        """
        return Reaction.from_fields(self.emoji, self.type)
    
    
    def delete_reaction_with(self, client):
        """
        Removes the added reaction.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will execute the action.
        
        Returns
        -------
        result : `int`
            The identifier number of the action what will be executed.
            
            Can be one of the following:
            +-----------------------+-------+
            | Respective name       | Value |
            +=======================+=======+
            | DELETE_REACTION_OK    | 0     |
            +-----------------------+-------+
            | DELETE_REACTION_PERM  | 1     |
            +-----------------------+-------+
        """
        if self.message.channel.cached_permissions_for(client) & PERMISSION_MASK_MANAGE_MESSAGES:
            Task(KOKORO, _delete_reaction_with_task(self, client))
            result = self.DELETE_REACTION_OK
        else:
            result = self.DELETE_REACTION_PERM
        
        return result



__all__ = ('Poll',)

from scarletio import RichAttributeErrorBaseType

from ...bases import maybe_snowflake
from ...utils import DATETIME_FORMAT_CODE

from ..poll_answer import PollAnswer
from ..poll_result import PollResult
from ..poll_result.utils import poll_result_sort_key

from .constants import DURATION_DEFAULT
from .fields import (
    parse_allow_multiple_choices, parse_answers, parse_duration, parse_expires_at, parse_finalized, parse_layout,
    parse_question, parse_results, put_allow_multiple_choices, put_answers, put_duration,
    put_expires_at, put_finalized, put_layout, put_question, put_results,
    validate_allow_multiple_choices, validate_answers, validate_duration, validate_expires_at, validate_finalized,
    validate_layout, validate_question, validate_results
)
from .preinstanced import PollLayout


class Poll(RichAttributeErrorBaseType):
    """
    Represents a poll.
    
    Attributes
    ----------
    allow_multiple_choices : `bool`
        The poll's allow_multiple_choices.
    answers : `None`, `tuple` of ``PollAnswer``
        The poll's answers.
    duration : `int`
        After how much time is the poll expiring. Should be multiple of 3600.
    expires_at : `None | DateTime`
        When the poll expires.
    finalized : `bool`
        Whether the poll is manually ended.
    layout : ``PollLayout``
        The poll's layout.
    question : `None`, ``PollQuestion``
        The poll's question.
    results : `None | list<PollResult>`
        The poll's results.
    """
    __slots__ = (
        'allow_multiple_choices', 'answers', 'duration', 'expires_at', 'finalized', 'layout', 'question', 'results'
    )
    
    def __new__(
        cls,
        *,
        allow_multiple_choices = ...,
        answers = ...,
        duration = ...,
        expires_at = ...,
        finalized = ...,
        layout = ...,
        question = ...,
        results = ...,
    ):
        """
        Creates a new poll with the given parameters.
        
        Parameters
        ----------
        allow_multiple_choices : `None`, `bool`, Optional (Keyword only)
            Whether selecting multiple answers is allowed.
        answers :  `None`, `iterable` of ``PollAnswer```, `iterable` of `str`, Optional (Keyword only)
            The poll's answers.
        duration : `int`, Optional (Keyword only)
            After how much time is the poll expiring. Should be multiple of 3600.
        expires_at : `None | DateTime`, Optional (Keyword only)
            When the poll expires.
        finalized : `bool`, Optional (Keyword only)
            Whether the poll is manually ended.
        layout : ``PollLayout``, `int`, Optional (Keyword only)
            The poll's layout.
        question : `None`, ``PollQuestion``, `str`, Optional (Keyword only)
            The poll's question.
        results : `None`, `iterable` of ``PollResult``, Optional (Keyword only)
            The poll's results.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # allow_multiple_choices
        if allow_multiple_choices is ...:
            allow_multiple_choices = False
        else:
            allow_multiple_choices = validate_allow_multiple_choices(allow_multiple_choices)
        
        # answers
        if answers is ...:
            answers = None
        else:
            answers = validate_answers(answers)
        
        # duration
        if duration is ...:
            duration = DURATION_DEFAULT
        else:
            duration = validate_duration(duration)
        
        # expires_at
        if expires_at is ...:
            expires_at = None
        else:
            expires_at = validate_expires_at(expires_at)
        
        # finalized
        if finalized is ...:
            finalized = False
        else:
            finalized = validate_finalized(finalized)
        
        # layout
        if layout is ...:
            layout = PollLayout.default
        else:
            layout = validate_layout(layout)
        
        # question
        if question is ...:
            question = None
        else:
            question = validate_question(question)
        
        # results
        if results is ...:
            results = None
        else:
            results = validate_results(results)
        
        # Construct
        self = object.__new__(cls)
        self.allow_multiple_choices = allow_multiple_choices
        self.answers = answers
        self.duration = duration
        self.expires_at = expires_at
        self.finalized = finalized
        self.layout = layout
        self.question = question
        self.results = results
        
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty poll.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.allow_multiple_choices = False
        self.answers = None
        self.duration = DURATION_DEFAULT
        self.expires_at = None
        self.finalized = False
        self.layout = PollLayout.default
        self.question = None
        self.results = None
        return self
    
    
    def __len__(self):
        """Returns the poll's length."""
        length = 0
        
        # allow_multiple_choices
        # does not count
        
        # answers
        for answer in self.iter_answers():
            length += len(answer)
        
        # duration
        # does not count
        
        # expires_at
        # does not count
        
        # finalized
        # does not count
        
        # layout
        # does not count
        
        # question
        question = self.question
        if (question is not None):
            length += len(question)
        
        # results
        # does not count
        
        return length
    
    
    def __bool__(self):
        """Returns whether the poll has any fields set."""
        # allow_multiple_choices
        if self.allow_multiple_choices:
            return True
        
        # answers
        if self.answers is not None:
            return True
        
        # duration
        if self.duration != DURATION_DEFAULT:
            return True
        
        # expires_at
        if self.expires_at is not None:
            return True
        
        # finalized
        if self.finalized:
            return True
        
        # layout
        layout = self.layout
        if layout is not PollLayout.none and layout is not PollLayout.default:
            return True
        
        # question
        if self.question is not None:
            return True
        
        # results
        # does not count
        return False
    
    
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # allow_multiple_choices
        allow_multiple_choices = self.allow_multiple_choices
        if allow_multiple_choices:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allow_multiple_choices = ')
            repr_parts.append(repr(allow_multiple_choices))
        
        # answers
        answers = self.answers
        if (answers is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' answers = ')
            repr_parts.append(repr(answers))
        
        # duration
        duration = self.duration
        if (duration != DURATION_DEFAULT):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' duration = ')
            repr_parts.append(repr(duration))
        
        # expires_at
        expires_at = self.expires_at
        if (expires_at is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' expires_at = ')
            repr_parts.append(format(expires_at, DATETIME_FORMAT_CODE))
        
        # finalized
        finalized = self.finalized
        if finalized:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' finalized = ')
            repr_parts.append(repr(finalized))
        
        # layout
        layout = self.layout
        if layout is not PollLayout.none and layout is not PollLayout.default:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' layout = ')
            repr_parts.append(layout.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(layout.value))
        
        # question
        question = self.question
        if (question is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' question = ')
            repr_parts.append(repr(question))
        
        # results
        results = self.results
        if (results is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' results = ')
            repr_parts.append(repr(results))
            
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the poll's hash value."""
        hash_value = 0
        
        # allow_multiple_choice
        hash_value ^= self.allow_multiple_choices << 8
        
        # answers
        answers = self.answers
        if (answers is not None):
            hash_value ^= len(answers)
            
            for answer in answers:
                hash_value ^= hash(answer)
        
        # duration
        duration = self.duration
        if (duration != DURATION_DEFAULT):
            hash_value ^= duration
        
        # expires_at
        expires_at = self.expires_at
        if (expires_at is not None):
            hash_value ^= hash(expires_at)
        
        # finalized
        hash_value ^= self.finalized << 12
        
        # layout
        hash_value ^= hash(self.layout)
        
        # question
        question = self.question
        if (question is not None):
            hash_value ^= hash(question)
        
        # results | order does not matter
        results = self.results
        if (results is not None):
            hash_value ^= len(results) << 4
            
            for result in results:
                hash_value ^= hash(result)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two polls are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # allow_multiple_choices
        if self.allow_multiple_choices != other.allow_multiple_choices:
            return False
        
        # answers
        if self.answers != other.answers:
            return False
        
        # duration
        if self.duration != other.duration:
            return False
        
        # expires_at
        if self.expires_at != other.expires_at:
            return False
        
        # finalized
        if self.finalized != other.finalized:
            return False
        
        # layout
        if self.layout is not other.layout:
            return False
        
        # question
        if self.question != other.question:
            return False
        
        # results | order matters
        self_results = self.results
        if (self_results is not None):
            self_results.sort(key = poll_result_sort_key)
        
        other_results = other.results
        if (other_results is not None):
            other_results.sort(key = poll_result_sort_key)
        
        if self_results != other_results:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new poll instance from the given json data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create poll from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.allow_multiple_choices = parse_allow_multiple_choices(data)
        self.answers = parse_answers(data)
        self.duration = parse_duration(data)
        self.expires_at = parse_expires_at(data)
        self.finalized = parse_finalized(data)
        self.layout = parse_layout(data)
        self.question = parse_question(data)
        self.results = parse_results(data)
        return self
    
    
    def _update_attributes(self, data):
        """
        Updates the poll's attributes with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to update the poll with.
        """
        self.allow_multiple_choices = parse_allow_multiple_choices(data)
        self.answers = parse_answers(data)
        self.duration = parse_duration(data)
        self.expires_at = parse_expires_at(data)
        self.finalized = parse_finalized(data)
        self.layout = parse_layout(data)
        self.question = parse_question(data)
        self.results = parse_results(data, self.results)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the poll's attributes with the given data  and returns the changes in a
        `dict` of (`attribute-name`, `old-value`) items.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to update the poll with.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------------------+-----------------------------------+
        | Keys                      | Values                            |
        +===========================+===================================+
        | allow_multiple_choices    | `bool`                            |
        +---------------------------+-----------------------------------+
        | answers                   | `None`, `tuple` of ``PollAnswer`` |
        +---------------------------+-----------------------------------+
        | duration                  | `int`                             |
        +---------------------------+-----------------------------------+
        | expires_at                | `None`, `DateTime`                |
        +---------------------------+-----------------------------------+
        | finalized                 | `bool`                            |
        +---------------------------+-----------------------------------+
        | layout                    | ``PollLayout``                    |
        +---------------------------+-----------------------------------+
        | question                  | `None`, ``PollQuestion``          |
        +---------------------------+-----------------------------------+
        """
        old_attributes = {}
        
        # allow_multiple_choices
        allow_multiple_choices = parse_allow_multiple_choices(data)
        if (allow_multiple_choices != self.allow_multiple_choices):
            old_attributes['allow_multiple_choices'] = self.allow_multiple_choices
            self.allow_multiple_choices = allow_multiple_choices
        
        # answers
        answers = parse_answers(data)
        if (answers != self.answers):
            old_attributes['answers'] = self.answers
            self.answers = answers
        
        # duration
        duration = parse_duration(data)
        if (duration != self.duration):
            old_attributes['duration'] = self.duration
            self.duration = duration
        
        # expires_at
        expires_at = parse_expires_at(data)
        if (expires_at != self.expires_at):
            old_attributes['expires_at'] = self.expires_at
            self.expires_at = expires_at
        
        # finalized
        finalized = parse_finalized(data)
        if (finalized != self.finalized):
            old_attributes['finalized'] = self.finalized
            self.finalized = finalized
        
        # layout
        layout = parse_layout(data)
        if (layout is not self.layout):
            old_attributes['layout'] = self.layout
            self.layout = layout
        
        # question
        question = parse_question(data)
        if (question != self.question):
            old_attributes['question'] = self.question
            self.question = question
        
        # results
        self.results = parse_results(data, self.results)
        
        return old_attributes
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Returns the poll as a json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_allow_multiple_choices(self.allow_multiple_choices, data, defaults)
        put_answers(self.answers, data, defaults, include_internals = include_internals)
        put_layout(self.layout, data, defaults)
        put_question(self.question, data, defaults)
        
        if include_internals:
            put_expires_at(self.expires_at, data, defaults)
            put_finalized(self.finalized, data, defaults)
            put_results(self.results, data, defaults)
        else:
            put_duration(self.duration, data, defaults)
        
        return data
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the poll by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : `None`, ``Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.allow_multiple_choices = self.allow_multiple_choices
        
        answers = self.answers
        if (answers is not None):
            answers = (*(answer.clean_copy(guild) for answer in answers),)
        new.answers = answers
        
        new.duration = self.duration
        new.expires_at = self.expires_at
        new.finalized = self.finalized
        new.layout = self.layout
        
        question = self.question
        if (question is not None):
            question = question.clean_copy(guild)
        new.question = question
        
        results = self.results
        if (results is not None):
            results = [result.copy() for result in results]
        new.results = results
        
        return new
    
    
    def copy(self):
        """
        Copies the poll.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.allow_multiple_choices = self.allow_multiple_choices
        
        answers = self.answers
        if (answers is not None):
            answers = (*(answer.copy() for answer in answers),)
        new.answers = answers
        
        new.duration = self.duration
        new.expires_at = self.expires_at
        new.finalized = self.finalized
        new.layout = self.layout
        
        question = self.question
        if (question is not None):
            question = question.copy()
        new.question = question
        
        results = self.results
        if (results is not None):
            results = [result.copy() for result in results]
        new.results = results
        
        return new
    
    
    def copy_with(
        self,
        *,
        allow_multiple_choices = ...,
        answers = ...,
        duration = ...,
        expires_at = ...,
        finalized = ...,
        layout = ...,
        question = ...,
        results = ...,
    ):
        """
        Copies the poll with the given parameters.
        
        Parameters
        ----------
        allow_multiple_choices : `bool`, Optional (Keyword only)
            Whether selecting multiple answers is allowed.
        answers :  `None`, `iterable` of ``PollAnswer```, `iterable` of `str`, Optional (Keyword only)
            The poll's answers.
        duration : `int`, Optional (Keyword only)
            After how much time is the poll expiring. Should be multiple of 3600.
        expires_at : `None | DateTime`, Optional (Keyword only)
            When the poll expires.
        finalized : `bool`, Optional (Keyword only)
            Whether the poll is manually ended.
        layout : ``PollLayout``, `int`, Optional (Keyword only)
            The poll's layout.
        question : `None`, ``PollQuestion``, `str`, Optional (Keyword only)
            The poll's question.
        results : `None`, `iterable` of ``PollResult``, Optional (Keyword only)
            The poll's results.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # allow_multiple_choices
        if allow_multiple_choices is ...:
            allow_multiple_choices = self.allow_multiple_choices
        else:
            allow_multiple_choices = validate_allow_multiple_choices(allow_multiple_choices)
        
        # answers
        if answers is ...:
            answers = self.answers
            if (answers is not None):
                answers = (*(answer.copy() for answer in answers),)
        else:
            answers = validate_answers(answers)
        
        # duration
        if duration is ...:
            duration = self.duration
        else:
            duration = validate_duration(duration)
        
        # expires_at
        if expires_at is ...:
            expires_at = self.expires_at
        else:
            expires_at = validate_expires_at(expires_at)
        
        # finalized
        if finalized is ...:
            finalized = self.finalized
        else:
            finalized = validate_finalized(finalized)
        
        # layout
        if layout is ...:
            layout = self.layout
        else:
            layout = validate_layout(layout)
        
        # question
        if question is ...:
            question = self.question
            if (question is not None):
                question = question.copy()
        else:
            question = validate_question(question)
        
        # results
        if results is ...:
            results = self.results
            if (results is not None):
                results = [result.copy() for result in results]
        else:
            results = validate_results(results)
        
        # Construct
        new = object.__new__(type(self))
        new.allow_multiple_choices = allow_multiple_choices
        new.answers = answers
        new.duration = duration
        new.expires_at = expires_at
        new.finalized = finalized
        new.layout = layout
        new.question = question
        new.results = results
        return new
    
    
    @property
    def contents(self):
        """
        Returns the contents of the poll.
        
        Returns
        -------
        contents : `list<str>`
        """
        return [*self.iter_contents()]
    
    
    def iter_contents(self):
        """
        Iterates over the contents of the poll.
        
        This method is an iterable generator.
        
        Yields
        ------
        content : `str`
        """
        question = self.question
        if (question is not None):
            yield from question.iter_contents()
        
        for answer in self.iter_answers():
            yield from answer.iter_contents()
    
    
    def iter_answers(self):
        """
        Iterates over the answers of the poll.
        
        This method is an iterable generator.
        
        Yields
        ------
        answer : ``PollAnswer``
        """
        answers = self.answers
        if (answers is not None):
            yield from answers
    
    
    def iter_results(self):
        """
        Iterates over the results of the poll.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``PollResult``
        """
        # Order does not matter.
        results = self.results
        if (results is not None):
            yield from self.results
    
    
    def get_answer_id(self, poll_answer):
        """
        Gets the poll answer's identifier for an answer like the given one.
        
        Parameters
        ----------
        poll_answer : ``PollAnswer``
            The poll answer to gets identifier of.
        
        Returns
        -------
        answer_id : `int`
            Returns `0` on failure.
        """
        answer_emoji = poll_answer.emoji
        answer_text = poll_answer.text
        for answer in self.iter_answers():
            if (answer_emoji is answer.emoji) and (answer_text == answer.text):
                return answer.id
        
        return 0
    
    
    def __getitem__(self, key):
        """
        Gets the result for the given answer.
        
        Parameters
        ----------
        key : ``PollAnswer``
            
        Returns
        -------
        result : `None | PollAnswer`
        
        Raises
        ------
        TypeError
        """
        if isinstance(key, PollAnswer):
            answer_id = key.id
            if not answer_id:
                answer_id = self.get_answer_id(key)
                if not answer_id:
                    return None
        
        else:
            answer_id = maybe_snowflake(key)
            if answer_id is None:
                raise TypeError(
                    f'`key` can be `int`, `{PollAnswer.__name__}`, got {type(key).__name__}; {key!r}.'
                )
        
        for result in self.iter_results():
            if result.answer_id == answer_id:
                return result
        
        return None
    
    
    def _get_or_create_result(self, answer_id):
        """
        Gets result for the given `answer_id`. If not found creates a new one.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        
        Returns
        -------
        result : ``PollResult``
        """
        results = self.results
        if results is None:
            result = PollResult._create_empty(answer_id)
            self.results = [result]
        else:
            for result in results:
                if answer_id == result.answer_id:
                    break
            
            else:
                result = PollResult._create_empty(answer_id)
                results.append(result)
        
        return result
    
    
    def _add_vote(self, answer_id, user):
        """
        Adds a vote on the poll.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        user : ``ClientUserBase``
            The user who voted.
        
        Returns
        -------
        success : `bool`
        """
        result = self._get_or_create_result(answer_id)
        return result._add_vote(user)
    
    
    def _fill_some_votes(self, answer_id, users):
        """
        Fills out some voter users for the given `answer_id` on the poll.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        users : `list` of ``ClientUserBase``
            The users who voted.
        """
        result = self._get_or_create_result(answer_id)
        result._fill_some_votes(users)
    
    
    def _fill_all_votes(self, answer_id, users):
        """
        Fills out all voter users for the given `answer_id` on the poll.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        users : `list` of ``ClientUserBase``
            The users who voted.
        """
        result = self._get_or_create_result(answer_id)
        result._fill_all_votes(users)
    
    
    def _remove_vote(self, answer_id, user):
        """
        Removes a vote from the poll.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        user : ``ClientUserBase``
            The user who removed their vote.
        
        Returns
        -------
        success : `bool`
        """
        results = self.results
        if results is None:
            return False
        
        for result in results:
            if answer_id == result.answer_id:
                break
        
        else:
            return False
        
        return result._remove_vote(user)
    
    
    def iter_items(self):
        """
        Iterates over `answer - result` pairs. If an answer does not have a result yet, creates it.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : `(PollAnswer, PollResult)`
        """
        for answer in self.iter_answers():
            yield answer, self._get_or_create_result(answer.id)



__all__ = ()

from ...field_parsers import nullable_entity_parser_factory
from ...field_putters import nullable_entity_optional_putter_factory
from ...field_validators import nullable_entity_validator_factory

from ..application_install_parameters import ApplicationInstallParameters

# install_parameters

parse_install_parameters = nullable_entity_parser_factory('oauth2_install_params', ApplicationInstallParameters)
put_install_parameters = nullable_entity_optional_putter_factory(
    'oauth2_install_params', ApplicationInstallParameters
)
validate_install_parameters = nullable_entity_validator_factory('install_parameters', ApplicationInstallParameters)



__all__ = ('FlagDeprecation', )

from datetime import datetime as DateTime, timezone as TimeZone
from warnings import warn

from scarletio import RichAttributeErrorBaseType


NOW = DateTime.now(tz = TimeZone.utc)
FORMAT_CODE = '%Y %B'


class FlagDeprecation(RichAttributeErrorBaseType):
    """
    Represents deprecation notice for a flag.
    
    Attributes
    ----------
    allowed : `bool`
        Whether the deprecation notice is allowed to trigger.
    
    removed_after : `str`
        When the field supposed to be removed.
    
    use_instead : `str`
        What should be used instead of this feature.
    """
    __slots__ = ('allowed', 'removed_after', 'use_instead')
    
    def __new__(cls, use_instead, removed_after, *, trigger_after = None):
        """
        Creates a new flag deprecation.
        
        Parameters
        ----------
        user_instance : `str`
            What should be used instead.
        
        removed_after : `DateTime`
            When the feature will be removed.
        
        trigger_after : `None | DateTime` = `None`, Optional (Keyword only)
            After when the deprecation should trigger.
        """
        # create allowed from trigger_after
        if trigger_after is None:
            allowed = True
        else:
            allowed = NOW > trigger_after
        
        # process removed_after
        removed_after = format(removed_after, FORMAT_CODE)
        
        # Construct
        self = object.__new__(cls)
        self.allowed = allowed
        self.removed_after = removed_after
        self.use_instead = use_instead
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two deprecations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # allowed
        if self.allowed != other.allowed:
            return False
        
        # removed_after
        if self.removed_after != other.removed_after:
            return False
        
        # use_instead
        if self.use_instead != other.use_instead:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the deprecation's hash value."""
        hash_value = 0
        
        # allowed
        hash_value ^= self.allowed
        
        # removed_after
        hash_value ^= hash(self.removed_after)
        
        # use_instead
        hash_value ^= hash(self.use_instead)
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the deprecation's representation"""
        repr_parts = ['<', type(self).__name__]
        
        # allowed
        repr_parts.append(', allowed = ')
        repr_parts.append(repr(self.allowed))
        
        # removed_after
        repr_parts.append(', removed_after = ')
        repr_parts.append(repr(self.removed_after))
        
        # use_instead
        repr_parts.append(', use_instead = ')
        repr_parts.append(repr(self.use_instead))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def trigger(self, type_name, flag_name, stack_level):
        """
        Triggers the deprecation.
        
        Parameters
        ----------
        type_name : `str`
            The deprecated's name.
        
        flag_name : `str`
            The deprecated flags name.
        
        stack_level : `int`
            The stack level to warn at.
        
        Returns
        -------
        warned : `bool`
        """
        if not self.allowed:
            return False
        
        warn(
            (
                f'`{type_name}.{flag_name}` is deprecated and will be removed at {self.removed_after}. '
                f'Please use {self.use_instead} instead.'
            ),
            FutureWarning,
            stacklevel = stack_level,
        )
        return True



__all__ = ('UserClan',)

from ...bases import IconSlot
from ...bases import Slotted
from ...core import GUILDS
from ...http import urls as module_urls

from .fields import (
    parse_enabled, parse_guild_id, parse_tag, put_enabled, put_guild_id, put_tag, validate_enabled,
    validate_guild_id, validate_tag
)


class UserClan(metaclass = Slotted):
    """
    Entity representing a server type ``Webhook``'s guild.
    
    Attributes
    ----------
    enabled : `bool`
        Whether the user is displaying their clan.
    guild_id : `int`
        The unique identifier number of the respective guild.
    icon_hash : `int`
        The clan's icon's hash in `uint128`.
    icon_type : ``IconType``
        The clan's icon's type.
    tag : `str`
        The clan's tag.
    """
    __slots__ = ('enabled', 'guild_id', 'tag')
    
    icon = IconSlot(
        'icon',
        'badge',
        module_urls.clan_icon_url,
        module_urls.clan_icon_url_as,
    )
    
    def __new__(cls, *, enabled = ..., guild_id = ..., icon = ..., tag = ...):
        """
        Creates a new user clan from the given fields.
        
        Parameters
        ----------
        enabled : `bool`, Optional (Keyword only)
            Whether the user is displaying their clan.
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier.
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The clan's icon.
        tag : `str`, Optional (Keyword only)
            The clan's tag.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # enabled
        if enabled is ...:
            enabled = True
        else:
            enabled = validate_enabled(enabled)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon)
        
        # tag
        if tag is ...:
            tag = ''
        else:
            tag = validate_tag(tag)
        
        # Constructor
        self = object.__new__(cls)
        self.enabled = enabled
        self.guild_id = guild_id
        self.icon = icon
        self.tag = tag
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new user clan from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to deserialize.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.enabled = parse_enabled(data)
        self.guild_id = parse_guild_id(data)
        self._set_icon(data)
        self.tag = parse_tag(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the user clan to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_enabled(self.enabled, data, defaults)
        put_guild_id(self.guild_id, data, defaults)
        type(self).icon.put_into(self.icon, data, defaults)
        put_tag(self.tag, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the user clan's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', tag = ')
        repr_parts.append(repr(self.tag))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the user clan's hash value."""
        hash_value = 0
        
        # enabled
        hash_value ^= self.enabled
        
        # guild_id
        hash_value ^= self.guild_id
        
        # icon
        hash_value ^= hash(self.icon)
        
        # tag
        hash_value ^= hash(self.tag)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two user clans are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # icon
        if self.icon_hash != other.icon_hash:
            return False
        if self.icon_type != other.icon_type:
            return False
        
        # tag
        if self.tag != other.tag:
            return False
        
        return True
    
    
    @property
    def guild(self):
        """
        Returns the guild of the user clan.
        
        If the guild is not cached returns `None`.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def copy(self):
        """
        Copies the user clan.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.enabled = self.enabled
        new.guild_id = self.guild_id
        new.icon = self.icon
        new.tag = self.tag
        return new
    
    
    def copy_with(self, *, enabled = ..., guild_id = ..., icon = ..., tag = ...):
        """
        Copies a new user clan from the given fields.
        
        Parameters
        ----------
        enabled : `bool`, Optional (Keyword only)
            Whether the user is displaying their clan.
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier.
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The clan's icon.
        tag : `str`, Optional (Keyword only)
            The clan's tag.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # enabled
        if enabled is ...:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon)
        
        # tag
        if tag is ...:
            tag = self.tag
        else:
            tag = validate_tag(tag)
        
        # Constructor
        new = object.__new__(type(self))
        new.enabled = enabled
        new.guild_id = guild_id
        new.icon = icon
        new.tag = tag
        return new



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory, int_parser_factory,
    int_postprocess_parser_factory, nullable_array_parser_factory, nullable_entity_array_parser_factory,
    nullable_entity_parser_factory, nullable_string_parser_factory, preinstanced_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    flag_optional_putter_factory, force_string_putter_factory, int_optional_postprocess_putter_factory,
    int_putter_factory, nullable_entity_array_putter_factory, nullable_entity_optional_putter_factory,
    nullable_string_array_optional_putter_factory, nullable_string_putter_factory, preinstanced_array_putter_factory,
    preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_entity_array_validator_factory, nullable_entity_validator_factory,
    nullable_string_array_validator_factory, nullable_string_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory, url_array_optional_validator_factory, url_optional_validator_factory
)
from ...guild import Guild
from ...user import ClientUserBase, User, ZEROUSER

from ..application_entity import ApplicationEntity
from ..application_executable import ApplicationExecutable
from ..application_install_parameters import ApplicationInstallParameters
from ..application_integration_type_configuration import ApplicationIntegrationTypeConfiguration
from ..embedded_activity_configuration import EmbeddedActivityConfiguration
from ..eula import EULA
from ..team import Team
from ..third_party_sku import ThirdPartySKU

from .constants import (
    BOT_PUBLIC_DEFAULT, BOT_REQUIRES_CODE_GRANT_DEFAULT, DESCRIPTION_LENGTH_MAX, HOOK_DEFAULT,
    INTEGRATION_PUBLIC_DEFAULT, INTEGRATION_REQUIRES_CODE_GRANT_DEFAULT, MAX_PARTICIPANTS_DEFAULT, NAME_LENGTH_MAX,
    NAME_LENGTH_MIN, OVERLAY_COMPATIBILITY_HOOK_DEFAULT, OVERLAY_DEFAULT
)
from .flags import (
    ApplicationDiscoveryEligibilityFlags, ApplicationFlag, ApplicationMonetizationEligibilityFlags,
    ApplicationOverlayMethodFlags
)
from .preinstanced import (
    ApplicationDiscoverabilityState, ApplicationEventWebhookEventType, ApplicationEventWebhookState,
    ApplicationExplicitContentFilterLevel, ApplicationIntegrationType, ApplicationInteractionEventType,
    ApplicationInteractionVersion, ApplicationInternalGuildRestriction, ApplicationMonetizationState,
    ApplicationRPCState, ApplicationStoreState, ApplicationType, ApplicationVerificationState
)

# aliases

parse_aliases = nullable_array_parser_factory('aliases')
put_aliases = nullable_string_array_optional_putter_factory('aliases')
validate_aliases = nullable_string_array_validator_factory('aliases')

# approximate_guild_count

parse_approximate_guild_count = int_parser_factory('approximate_guild_count', 0)
put_approximate_guild_count = int_putter_factory('approximate_guild_count')
validate_approximate_guild_count = int_conditional_validator_factory(
    'approximate_guild_count',
    0,
    (lambda approximate_guild_count : approximate_guild_count >= 0),
    '>= 0',
)

# approximate_user_install_count

parse_approximate_user_install_count = int_parser_factory('approximate_user_install_count', 0)
put_approximate_user_install_count = int_putter_factory('approximate_user_install_count')
validate_approximate_user_install_count = int_conditional_validator_factory(
    'approximate_user_install_count',
    0,
    (lambda approximate_user_install_count : approximate_user_install_count >= 0),
    '>= 0',
)

# bot_public

parse_bot_public = bool_parser_factory('bot_public', BOT_PUBLIC_DEFAULT)
put_bot_public = bool_optional_putter_factory('bot_public', BOT_PUBLIC_DEFAULT)
validate_bot_public = bool_validator_factory('bot_public', BOT_PUBLIC_DEFAULT)

# bot_requires_code_grant

parse_bot_requires_code_grant = bool_parser_factory('bot_require_code_grant', BOT_REQUIRES_CODE_GRANT_DEFAULT)
put_bot_requires_code_grant = bool_optional_putter_factory('bot_require_code_grant', BOT_REQUIRES_CODE_GRANT_DEFAULT)
validate_bot_requires_code_grant = bool_validator_factory('bot_requires_code_grant', BOT_REQUIRES_CODE_GRANT_DEFAULT)

# creator_monetization_state

parse_creator_monetization_state = preinstanced_parser_factory(
    'creator_monetization_state', ApplicationMonetizationState, ApplicationMonetizationState.none
)
put_creator_monetization_state = preinstanced_putter_factory('creator_monetization_state')
validate_creator_monetization_state = preinstanced_validator_factory(
    'creator_monetization_state', ApplicationMonetizationState
)

# custom_install_url

parse_custom_install_url = nullable_string_parser_factory('custom_install_url')
put_custom_install_url = url_optional_putter_factory('custom_install_url')
validate_custom_install_url = url_optional_validator_factory('custom_install_url')

# deeplink_url

parse_deeplink_url = nullable_string_parser_factory('deeplink_uri')
put_deeplink_url = url_optional_putter_factory('deeplink_uri')
validate_deeplink_url = url_optional_validator_factory('deeplink_url')

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# developers

parse_developers = nullable_entity_array_parser_factory('developers', ApplicationEntity)
put_developers = nullable_entity_array_putter_factory('developers', ApplicationEntity)
validate_developers = nullable_entity_array_validator_factory('developers', ApplicationEntity)

# discoverability_state

parse_discoverability_state = preinstanced_parser_factory(
    'discoverability_state', ApplicationDiscoverabilityState, ApplicationDiscoverabilityState.none
)
put_discoverability_state = preinstanced_putter_factory('discoverability_state')
validate_discoverability_state = preinstanced_validator_factory(
    'discoverability_state', ApplicationDiscoverabilityState
)


# discovery_eligibility_flags

parse_discovery_eligibility_flags = flag_parser_factory(
    'discovery_eligibility_flags', ApplicationDiscoveryEligibilityFlags
)
put_discovery_eligibility_flags = flag_optional_putter_factory(
    'discovery_eligibility_flags', ApplicationDiscoveryEligibilityFlags()
)
validate_discovery_eligibility_flags = flag_validator_factory(
    'discovery_eligibility_flags', ApplicationDiscoveryEligibilityFlags
)


# embedded_activity_configuration

parse_embedded_activity_configuration = nullable_entity_parser_factory(
    'embedded_activity_config', EmbeddedActivityConfiguration
)
put_embedded_activity_configuration = nullable_entity_optional_putter_factory(
    'embedded_activity_config', EmbeddedActivityConfiguration
)
validate_embedded_activity_configuration = nullable_entity_validator_factory(
    'embedded_activity_configuration', EmbeddedActivityConfiguration
)


# eula_id

parse_eula_id = entity_id_parser_factory('eula_id')
put_eula_id = entity_id_optional_putter_factory('eula_id')
validate_eula_id = entity_id_validator_factory('eula_id', EULA)


# event_webhook_event_types

parse_event_webhook_event_types = preinstanced_array_parser_factory(
    'event_webhooks_types', ApplicationEventWebhookEventType
)
put_event_webhook_event_types = preinstanced_array_putter_factory('event_webhooks_types')
validate_event_webhook_event_types = preinstanced_array_validator_factory(
    'event_webhook_event_types', ApplicationEventWebhookEventType
)


# event_webhook_state

parse_event_webhook_state = preinstanced_parser_factory(
    'event_webhooks_status', ApplicationEventWebhookState, ApplicationEventWebhookState.none
)
put_event_webhook_state = preinstanced_putter_factory('event_webhooks_status')
validate_event_webhook_state = preinstanced_validator_factory(
    'event_webhook_state', ApplicationEventWebhookState
)


# event_webhook_url

parse_event_webhook_url = nullable_string_parser_factory('event_webhooks_url')
put_event_webhook_url = url_optional_putter_factory('event_webhooks_url')
validate_event_webhook_url = url_optional_validator_factory('event_webhook_url')


# executables

parse_executables = nullable_entity_array_parser_factory('executables', ApplicationExecutable)
put_executables = nullable_entity_array_putter_factory('executables', ApplicationExecutable)
validate_executables = nullable_entity_array_validator_factory('executables', ApplicationExecutable)


# explicit_content_filter_level

parse_explicit_content_filter_level = preinstanced_parser_factory(
    'explicit_content_filter', ApplicationExplicitContentFilterLevel, ApplicationExplicitContentFilterLevel.none
)
put_explicit_content_filter_level = preinstanced_putter_factory('explicit_content_filter')
validate_explicit_content_filter_level = preinstanced_validator_factory(
    'explicit_content_filter_level', ApplicationExplicitContentFilterLevel
)


# flags

parse_flags = flag_parser_factory('flags', ApplicationFlag)
put_flags = flag_optional_putter_factory('flags', ApplicationFlag())
validate_flags = flag_validator_factory('flags', ApplicationFlag)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# hook

parse_hook = bool_parser_factory('hook', HOOK_DEFAULT)
put_hook = bool_optional_putter_factory('hook', HOOK_DEFAULT)
validate_hook = bool_validator_factory('hook', HOOK_DEFAULT)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('application_id')

# install_parameters

parse_install_parameters = nullable_entity_parser_factory('install_params', ApplicationInstallParameters)
put_install_parameters = nullable_entity_optional_putter_factory('install_params', ApplicationInstallParameters)
validate_install_parameters = nullable_entity_validator_factory('install_parameters', ApplicationInstallParameters)


# integration_public

parse_integration_public = bool_parser_factory('integration_public', INTEGRATION_PUBLIC_DEFAULT)
put_integration_public = bool_optional_putter_factory('integration_public', INTEGRATION_PUBLIC_DEFAULT)
validate_integration_public = bool_validator_factory('integration_public', INTEGRATION_PUBLIC_DEFAULT)


# integration_requires_code_grant

parse_integration_requires_code_grant = bool_parser_factory(
    'integration_require_code_grant', INTEGRATION_REQUIRES_CODE_GRANT_DEFAULT
)
put_integration_requires_code_grant = bool_optional_putter_factory(
    'integration_require_code_grant', INTEGRATION_REQUIRES_CODE_GRANT_DEFAULT
)
validate_integration_requires_code_grant = bool_validator_factory(
    'integration_requires_code_grant', INTEGRATION_REQUIRES_CODE_GRANT_DEFAULT
)

# integration_types

parse_integration_types = preinstanced_array_parser_factory('integration_types', ApplicationIntegrationType)
put_integration_types = preinstanced_array_putter_factory('integration_types')
validate_integration_types = preinstanced_array_validator_factory('integration_types', ApplicationIntegrationType)


# integration_types_configuration

def parse_integration_types_configuration(data):
    """
    Parses application integration types configuration.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    integration_types_configuration : `None | dict<ApplicationIntegrationType, ApplicationIntegrationTypeConfiguration>`
    """
    configurations_data = data.get('integration_types_config', None)
    if (configurations_data is None) or (not configurations_data):
        return None
    
    integration_types_configuration = {}
    for key, value in configurations_data.items():
        integration_type = ApplicationIntegrationType(ApplicationIntegrationType.VALUE_TYPE(key))
        integration_type_configuration = ApplicationIntegrationTypeConfiguration.from_data(value)
        integration_types_configuration[integration_type] = integration_type_configuration
    
    return integration_types_configuration


def put_integration_types_configuration(integration_types_configuration, data, defaults):
    """
    Puts the application's owner data into the given `data` json serializable object.
    
    Parameters
    ----------
    integration_types_configuration : `None | dict<ApplicationIntegrationType, ApplicationIntegrationTypeConfiguration>`
        Integration types configuration to serialize.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    configurations_data = {}
    
    if (integration_types_configuration is not None):
        for integration_type, integration_type_configuration in integration_types_configuration.items():
            key = str(integration_type.value)
            value = integration_type_configuration.to_data(defaults = defaults)
            configurations_data[key] = value 
    
    data['integration_types_config'] = configurations_data
    return data


def validate_integration_types_configuration(integration_types_configuration):
    """
    Validates the given `integration_types_configuration` value.
    
    Parameters
    ----------
    integration_types_configuration : \
            `None | dict<ApplicationIntegrationType | int, ApplicationIntegrationTypeConfiguration>`
        Integration types configuration to validate.
    
    Returns
    -------
    integration_types_configuration : `None | dict<ApplicationIntegrationType, ApplicationIntegrationTypeConfiguration>`
    
    Raises
    ------
    TypeError
        - If `integration_types_configuration`'s type is invalid.
    """
    if integration_types_configuration is None:
        return None
    
    if not isinstance(integration_types_configuration, dict):
        raise TypeError(
            f'`integration_types_configuration` can be `None`,'
            f'`dict<{ApplicationIntegrationType.__name__} | {ApplicationIntegrationType.VALUE_TYPE.__name__}, '
            f'{ApplicationIntegrationTypeConfiguration.__name__}`, got '
            f'{type(integration_types_configuration).__name__}; {integration_types_configuration!r}.'
        )
    
    if not integration_types_configuration:
        return None
    
    validated_integration_types_configuration = {}
    
    for key, value in integration_types_configuration.items():
        if isinstance(key, ApplicationIntegrationType):
            integration_type = key
        
        elif isinstance(key, ApplicationIntegrationType.VALUE_TYPE):
            integration_type = ApplicationIntegrationType(key)
        
        else:
            raise TypeError(
                f'`integration_types_configuration` keys can be '
                f'`{ApplicationIntegrationType.__name__}`, `{ApplicationIntegrationType.VALUE_TYPE.__name__}`, '
                f'got {type(key).__name__}; {key!r}; '
                f'integration_types_configuration = {integration_types_configuration!r}.'
            )
        
        if isinstance(value, ApplicationIntegrationTypeConfiguration):
            integration_type_configuration = value
        
        else:
            raise TypeError(
                f'`integration_types_configuration` values can be `{ApplicationIntegrationTypeConfiguration.__name__}`, '
                f'got {type(value).__name__}; {value!r}; '
                f'integration_types_configuration = {integration_types_configuration!r}.'
            )
        
        validated_integration_types_configuration[integration_type] = integration_type_configuration
    
    return validated_integration_types_configuration


# interaction_endpoint_url

parse_interaction_endpoint_url = nullable_string_parser_factory('interactions_endpoint_url')
put_interaction_endpoint_url = url_optional_putter_factory('interactions_endpoint_url')
validate_interaction_endpoint_url = url_optional_validator_factory('interaction_endpoint_url')


# interaction_event_types

parse_interaction_event_types = preinstanced_array_parser_factory(
    'interactions_event_types', ApplicationInteractionEventType
)
put_interaction_event_types = preinstanced_array_putter_factory('interactions_event_types')
validate_interaction_event_types = preinstanced_array_validator_factory(
    'interaction_event_types', ApplicationInteractionEventType
)

# interaction_version

parse_interaction_version = preinstanced_parser_factory(
    'interactions_version', ApplicationInteractionVersion, ApplicationInteractionVersion.none
)
put_interaction_version = preinstanced_putter_factory('interactions_version')
validate_interaction_version = preinstanced_validator_factory('interaction_version', ApplicationInteractionVersion)


# internal_guild_restriction

parse_internal_guild_restriction = preinstanced_parser_factory(
    'internal_guild_restriction', ApplicationInternalGuildRestriction, ApplicationInternalGuildRestriction.none
)
put_internal_guild_restriction = preinstanced_putter_factory('internal_guild_restriction')
validate_internal_guild_restriction = preinstanced_validator_factory(
    'internal_guild_restriction', ApplicationInternalGuildRestriction
)

# max_participants

parse_max_participants = int_postprocess_parser_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,
    (lambda max_participants: MAX_PARTICIPANTS_DEFAULT if max_participants == -1 else max_participants),
)
put_max_participants = int_optional_postprocess_putter_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,   
    (lambda max_participants: -1 if max_participants == MAX_PARTICIPANTS_DEFAULT else max_participants),
)
validate_max_participants = int_conditional_validator_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,
    (lambda width : width >= 0),
    '>= 0',
)


# monetization_eligibility_flags

parse_monetization_eligibility_flags = flag_parser_factory(
    'monetization_eligibility_flags', ApplicationMonetizationEligibilityFlags
)
put_monetization_eligibility_flags = flag_optional_putter_factory(
    'monetization_eligibility_flags', ApplicationMonetizationEligibilityFlags()
)
validate_monetization_eligibility_flags = flag_validator_factory(
    'monetization_eligibility_flags', ApplicationMonetizationEligibilityFlags
)


# monetization_state

parse_monetization_state = preinstanced_parser_factory(
    'monetization_state', ApplicationMonetizationState, ApplicationMonetizationState.none
)
put_monetization_state = preinstanced_putter_factory('monetization_state')
validate_monetization_state = preinstanced_validator_factory(
    'monetization_state', ApplicationMonetizationState
)


# monetized

parse_monetized = bool_parser_factory('is_monetized', False)
put_monetized = bool_optional_putter_factory('is_monetized', False)
validate_monetized = bool_validator_factory('monetized', False)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# overlay

parse_overlay = bool_parser_factory('overlay', OVERLAY_DEFAULT)
put_overlay = bool_optional_putter_factory('overlay', OVERLAY_DEFAULT)
validate_overlay = bool_validator_factory('overlay', OVERLAY_DEFAULT)

# overlay_compatibility_hook

parse_overlay_compatibility_hook = bool_parser_factory('overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT)
put_overlay_compatibility_hook = bool_optional_putter_factory(
    'overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT
)
validate_overlay_compatibility_hook = bool_validator_factory(
    'overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT
)

# overlay_method_flags

parse_overlay_method_flags = flag_parser_factory('overlay_methods', ApplicationOverlayMethodFlags)
put_overlay_method_flags = flag_optional_putter_factory('overlay_methods', ApplicationOverlayMethodFlags())
validate_overlay_method_flags = flag_validator_factory('overlay_method_flags', ApplicationOverlayMethodFlags)

# owner

def parse_owner(data):
    """
    Parses the application owner.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Application data.
    
    Returns
    -------
    owner : ``ClientUserBase``, ``Team``
    """
    team_data = data.get('team', None)
    if team_data is None:
        user_data = data.get('owner', None)
        if user_data is None:
            owner = ZEROUSER
        else:
            owner = User.from_data(user_data)
    else:
        owner = Team.from_data(team_data)
    
    return owner


def put_owner(owner, data, defaults):
    """
    Puts the application's owner data into the given `data` json serializable object.
    
    Parameters
    ----------
    owner : ``ClientUserBase``, ``Team``
        The application's owner.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if owner is ZEROUSER:
        team_data = None
        user_data = None
    
    elif isinstance(owner, ClientUserBase):
        team_data = None
        user_data = owner.to_data(defaults = defaults, include_internals = True)
        
    elif isinstance(owner, Team):
        team_data = owner.to_data(defaults = defaults, include_internals = True)
        user_data = owner.to_data_user()
    
    else:
        team_data = None
        user_data = None
    
    if (team_data is not None) or defaults:
        data['team'] = team_data
    
    if (user_data is not None) or defaults:
        data['owner'] = user_data
    
    return data


def validate_owner(owner):
    """
    Validates the given application owner.
    
    Parameters
    ----------
    owner : `None`, ``ClientUserBase``, ``Team``
        The application's owner.
    
    Returns
    -------
    owner : ``ClientUserBase``, ``Team``
    
    Raises
    ------
    TypeError
        - If `owner`'s type is incorrect.
    """
    if owner is None:
        owner = ZEROUSER
    
    elif not isinstance(owner, (ClientUserBase, Team)):
        raise TypeError(
            f'`owner` can be `None`, `{ClientUserBase.__name__}`, {Team.__name__}, '
            f'got {type(owner).__name__}; {owner!r}'
        )
    
    return owner

# primary_sku_id

parse_primary_sku_id = entity_id_parser_factory('primary_sku_id')
put_primary_sku_id = entity_id_optional_putter_factory('primary_sku_id')
validate_primary_sku_id = entity_id_validator_factory('primary_sku_id')

# privacy_policy_url

parse_privacy_policy_url = nullable_string_parser_factory('privacy_policy_url')
put_privacy_policy_url = url_optional_putter_factory('privacy_policy_url')
validate_privacy_policy_url = url_optional_validator_factory('privacy_policy_url')

# publishers

parse_publishers = nullable_entity_array_parser_factory('publishers', ApplicationEntity)
put_publishers = nullable_entity_array_putter_factory('publishers', ApplicationEntity)
validate_publishers = nullable_entity_array_validator_factory('publishers', ApplicationEntity)

# redirect_urls

def parse_redirect_urls(data):
    """
    Parses out `redirect_urls` value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    output : `None | tuple<str>`
    """
    string_array = data.get('redirect_uris', None)
    if string_array is None:
        return None
    
    parsed_values = None
    
    for string in string_array:
        if string is None:
            continue
        
        if parsed_values is None:
            parsed_values = []
        
        parsed_values.append(string)
    
    if parsed_values is None:
        return None
    
    parsed_values.sort()
    return tuple(parsed_values)


put_redirect_urls = nullable_string_array_optional_putter_factory('redirect_uris')
validate_redirect_urls = nullable_string_array_validator_factory('redirect_urls')
 

# role_connection_verification_url

parse_role_connection_verification_url = nullable_string_parser_factory('role_connections_verification_url')
put_role_connection_verification_url = url_optional_putter_factory('role_connections_verification_url')
validate_role_connection_verification_url = url_optional_validator_factory('role_connection_verification_url')

# rpc_origins

parse_rpc_origins = nullable_array_parser_factory('rpc_origins')
put_rpc_origins = nullable_string_array_optional_putter_factory('rpc_origins')
validate_rpc_origins = url_array_optional_validator_factory('rpc_origins')


# rpc_state

parse_rpc_state = preinstanced_parser_factory('rpc_application_state', ApplicationRPCState, ApplicationRPCState.none)
put_rpc_state = preinstanced_putter_factory('rpc_application_state')
validate_rpc_state = preinstanced_validator_factory('rpc_state', ApplicationRPCState)

# slug

parse_slug = nullable_string_parser_factory('slug')
put_slug = url_optional_putter_factory('slug')
validate_slug = url_optional_validator_factory('slug')

# store_state

parse_store_state = preinstanced_parser_factory(
    'store_application_state', ApplicationStoreState, ApplicationStoreState.none
)
put_store_state = preinstanced_putter_factory('store_application_state')
validate_store_state = preinstanced_validator_factory('store_state', ApplicationStoreState)

# tags

parse_tags = nullable_array_parser_factory('tags')
put_tags = nullable_string_array_optional_putter_factory('tags')
validate_tags = nullable_string_array_validator_factory('tags')

# terms_of_service_url

parse_terms_of_service_url = nullable_string_parser_factory('terms_of_service_url')
put_terms_of_service_url = url_optional_putter_factory('terms_of_service_url')
validate_terms_of_service_url = url_optional_validator_factory('terms_of_service_url')

# third_party_skus

parse_third_party_skus = nullable_entity_array_parser_factory('third_party_skus', ThirdPartySKU)
put_third_party_skus = nullable_entity_array_putter_factory('third_party_skus', ThirdPartySKU)
validate_third_party_skus = nullable_entity_array_validator_factory('third_party_skus', ThirdPartySKU)

# type

parse_type = preinstanced_parser_factory('type', ApplicationType, ApplicationType.none)


def put_type(application_type, data, defaults):
    """
    Puts the application's type into the given `data` json serializable object.
    
    Parameters
    ----------
    application_type : ``ApplicationType``
        The application's type.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if application_type is ApplicationType.none:
        value = None
    else:
        value = application_type.value
    
    data['type'] = value
    return data


validate_type = preinstanced_validator_factory('application_type', ApplicationType)


# verification_state

parse_verification_state = preinstanced_parser_factory(
    'verification_state', ApplicationVerificationState, ApplicationVerificationState.none
)
put_verification_state = preinstanced_putter_factory('verification_state')
validate_verification_state = preinstanced_validator_factory('verification_state', ApplicationVerificationState)


# verify_key

parse_verify_key = nullable_string_parser_factory('verify_key')
put_verify_key = nullable_string_putter_factory('verify_key')
validate_verify_key = nullable_string_validator_factory('verify_key', 0, 1024)



__all__ = ('MessageReferenceType',)

from ...bases import Preinstance as P, PreinstancedBase


class MessageReferenceType(PreinstancedBase, value_type = int):
    """
    Represents an message reference's type.
    
    Attributes
    ----------
    name : `str`
        The name of the message reference type.
    
    value : `int`
        The Discord side identifier value of the message reference type.
        
    Type Attributes
    ---------------
    Every predefined message reference type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | reply                     | reply                     | 0     |
    +---------------------------+---------------------------+-------+
    | forward                   | forward                   | 1     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    reply = P(0, 'reply')
    forward = P(1, 'forward')



__all__ = ('ScheduledEventEntityMetadataStage',)

from scarletio import copy_docs

from .base import ScheduledEventEntityMetadataBase
from .fields import parse_speaker_ids, put_speaker_ids, validate_speaker_ids


class ScheduledEventEntityMetadataStage(ScheduledEventEntityMetadataBase):
    """
    Stage entity metadata of ``ScheduledEvent``-s.
    
    Attributes
    ----------
    speaker_ids : `None`, `tuple` of `int`
        The speakers' identifier of the stage channel.
    """
    __slots__ = ('speaker_ids', )
    
    def __new__(cls, *, speaker_ids = ...):
        """
        Creates a new entity metadata instance.
        
        Parameters
        ----------
        speaker_ids : `None`, `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The speakers' identifier of the stage channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # speaker_ids
        if speaker_ids is ...:
            speaker_ids = None
        else:
            speaker_ids = validate_speaker_ids(speaker_ids)
        
        self = object.__new__(cls)
        self.speaker_ids = speaker_ids
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            speaker_ids = keyword_parameters.pop('speaker_ids', ...),
        )
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.speaker_ids = parse_speaker_ids(data)
        return self
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_speaker_ids(self.speaker_ids, data, defaults)
        return data
    
    
    @classmethod
    def _create_empty(cls):
        self = object.__new__(cls)
        self.speaker_ids = None
        return self
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            repr_parts.append(' speaker_ids = [')
            
            index = 0
            length = len(speaker_ids)
            
            while True:
                speaker_id = speaker_ids[index]
                repr_parts.append(repr(speaker_id))
                
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            hash_value ^= len(speaker_ids)
            
            for speaker_id in speaker_ids:
                hash_value ^= speaker_id
        
        return hash_value
    
    
    @copy_docs(ScheduledEventEntityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.speaker_ids != other.speaker_ids:
            return False
        
        return True
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            speaker_ids = (*speaker_ids,)
        new.speaker_ids = speaker_ids
        return new
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.copy_with)
    def copy_with(self, *, speaker_ids = ...):
        """
        Copies the scheduled event entity metadata with the given fields.
        
        Parameters
        ----------
        speaker_ids : `None`, `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The speakers' identifier of the stage channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # speaker_ids
        if speaker_ids is ...:
            speaker_ids = self.speaker_ids
            if (speaker_ids is not None):
                speaker_ids = (*speaker_ids,)
        else:
            speaker_ids = validate_speaker_ids(speaker_ids)
        
        new = object.__new__(type(self))
        new.speaker_ids = speaker_ids
        return new
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            speaker_ids = keyword_parameters.pop('speaker_ids', ...),
        )



__all__ = ('LIBRARY_CALLED_DIRECTLY', 'REGISTERED_COMMANDS', 'REGISTERED_COMMANDS_BY_NAME',)

import sys
from sys import modules
from os import getcwd as get_current_work_directory
from os.path import basename as get_file_name, dirname as get_directory_name, join as join_paths

from ... import __file__ as PACKAGE_INIT_FILE, __package__ as PACKAGE_NAME


UPPER_DIRECTORY = get_directory_name(get_directory_name(PACKAGE_INIT_FILE))
PACKAGE_MAIN_FILE = join_paths(get_directory_name(PACKAGE_INIT_FILE), '__main__.py')
WORKING_DIRECTORY = get_current_work_directory()
COMMAND_DIRECTORY = join_paths(UPPER_DIRECTORY, PACKAGE_NAME, 'main', 'commands')
EXTERNAL_IMPORT_ROUTES_FILE = join_paths(UPPER_DIRECTORY, PACKAGE_NAME, 'main', '.external')

COMMAND_IMPORT_ROUTE = (PACKAGE_NAME, 'main', 'commands')

PYTHON_FILE_POSTFIX_NAMES = frozenset(('.py', '.pyd', '.pyc', '.so'))
SYSTEM_DEFAULT_PARAMETER = 'help'

REGISTERED_COMMANDS = set()
REGISTERED_COMMANDS_BY_NAME = {}

# Get which file was executed.
main_module = modules.get('__main__', None)
if main_module is None:
    LIBRARY_CALLED_DIRECTLY = False
else:
    main_module_spec = main_module.__spec__
    if main_module_spec is not None:
        LIBRARY_CALLED_DIRECTLY = (main_module.__spec__.origin == PACKAGE_MAIN_FILE)
    
    else:
        executed_file = next(iter(sys.argv), None)
        if (executed_file is not None):
            executed_file = get_file_name(executed_file)
        if (UPPER_DIRECTORY != WORKING_DIRECTORY) and (executed_file == PACKAGE_NAME):
            LIBRARY_CALLED_DIRECTLY = True
        else:
            LIBRARY_CALLED_DIRECTLY = False

# Clear up
main_module = None
main_module_spec = None
executed_file = None

del main_module
del main_module_spec
del executed_file
